import os
import sys

import numpy as np
import torch

from esp_ppq.core import (
    GRU_QUANT_EXPONENT,
    LSTM_QUANT_EXPONENT,
    DataType,
    OperationQuantizationConfig,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    TargetPlatform,
    TensorQuantizationConfig,
    convert_any_to_numpy,
)
from esp_ppq.executor.base import OPERATION_FORWARD_TABLE
from esp_ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.log import NaiveLogger
from esp_ppq.parser.espdl.espdl_graph_utils import (
    fuse_downstream_operation,
    insert_concat_node,
    insert_dequantize_node,
    insert_quantize_node,
    insert_requantize_node,
    insert_reshape_node,
    insert_slice_node,
    insert_transpose_node,
)
from esp_ppq.parser.espdl.espdl_typedef import (
    ACTIVATION_OP_SET,
    MATH_OP_SET,
    QUANT_EXCLUDE_OP_SET,
    QUANT_OP_SET,
    EspQuantType,
    ExporterPatternInfo,
    LayoutAnnotation,
)
from esp_ppq.quantization.qfunction.linear import PPQLinearQuant_toInt

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))

logger = NaiveLogger.get_logger("ESPDL")


class EspdlQuantHelper:
    """Helper class for processing onnx qdq format"""

    @staticmethod
    def TQC_Exportable_Check(TQC: TensorQuantizationConfig, bounded_var: Variable) -> bool:
        if not TQC.can_export(True):
            logger.info(f"Skip {bounded_var.name} because it's not exportable")
            return False

        if TQC.visibility == QuantizationVisibility.INTERNAL:
            logger.info(f"Skip {bounded_var.name} because TAC visibility is internal")
            return False

        if TQC.num_of_bits == 8 and TQC.policy.has_property(QuantizationProperty.LINEAR):
            if TQC.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                range_check = TQC.quant_max <= 255 and TQC.quant_min >= 0
            else:
                range_check = TQC.quant_max <= 127 and TQC.quant_min >= -128
        else:
            range_check = True

        if not range_check:
            logger.warning(
                f"Is it not safe to export TQC({bounded_var.name}) to Onnx, "
                f"INT8 value range must be [-128, 127] or [0, 255], "
                f"however [{TQC.quant_min, TQC.quant_max}] was given."
            )
            return False
        return True


class InsertQuantTypePattern(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.platform in [TargetPlatform.ESPDL_INT8, TargetPlatform.ESPDL_S3_INT8, TargetPlatform.ESPDL_C_INT8]:
            op.attributes["quant_type"] = EspQuantType.S8
        elif op.platform in [
            TargetPlatform.ESPDL_INT16,
            TargetPlatform.ESPDL_S3_INT16,
            TargetPlatform.ESPDL_H_PRE_INT16,
            TargetPlatform.ESPDL_S3_H_PRE_INT16,
            TargetPlatform.ESPDL_C_INT16,
            TargetPlatform.ESPDL_C_H_PRE_INT16,
        ]:
            op.attributes["quant_type"] = EspQuantType.S16
        else:
            op.attributes["quant_type"] = EspQuantType.F32

        return op


class InsertQuantNodePattern(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var.source_op:
                    if var.source_op.type in QUANT_OP_SET:
                        assert var.source_op.num_of_input == 3, "Quantize Node Format Error, need as least 3 inputs."
                        assert isinstance(var.source_op, Operation)
                        continue
                    elif var in op.inputs:
                        if (
                            not isinstance(var.source_op, QuantableOperation)
                            or var.source_op.output_quant_config[var.source_op.outputs.index(var)].state
                            == QuantizationStates.FP32
                        ) and var.source_op.type not in QUANT_EXCLUDE_OP_SET:
                            logger.debug(f"Insert Quantize Node for {op.name}:{var.name}")
                            insert_quantize_node(
                                graph=graph,
                                var=inserting_var,
                                config=config,
                                op=inserting_op,
                            )
        return op


class InsertRequantNodePattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var.source_op:
                    if var.source_op.type in QUANT_OP_SET:
                        assert var.source_op.num_of_input == 3, "Quantize Node Format Error, need as least 3 inputs."
                        assert isinstance(var.source_op, Operation)
                        continue
                    elif var in op.inputs and isinstance(var.source_op, QuantableOperation):
                        source_op_output_var_index = var.source_op.outputs.index(var)
                        source_op_output_config = var.source_op.output_quant_config[source_op_output_var_index]
                        scale_diff = torch.max(torch.abs(source_op_output_config.scale - config.scale)).item()
                        zeropoint_diff = torch.max(torch.abs(source_op_output_config.offset - config.offset)).item()

                        if (
                            source_op_output_config.num_of_bits != config.num_of_bits
                            or scale_diff >= 1e-5
                            or zeropoint_diff >= 1e-1
                        ):
                            # if config
                            logger.debug(f"Insert Requantize Node for {op.name}:{var.name}")
                            insert_requantize_node(
                                graph=graph,
                                var=inserting_var,
                                upstream_config=source_op_output_config,
                                config=config,
                                op=inserting_op,
                            )

        return op


class InsertDequantNodePattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var in op.outputs:
                    for dest_op in var.dest_ops:
                        if (
                            dest_op
                            and dest_op.type not in QUANT_OP_SET
                            and not isinstance(dest_op, QuantableOperation)
                            and dest_op.type not in QUANT_EXCLUDE_OP_SET
                        ):
                            logger.debug(f"Insert Dequantize Node for {op.name}:{var.name}")
                            insert_dequantize_node(
                                graph=graph,
                                var=inserting_var,
                                config=config,
                                op=dest_op,
                            )
        return op


class InsertPreNodeOfMatMulPattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type != "MatMul" or op.inputs[1].is_parameter or not isinstance(op, QuantableOperation):
            return op

        input0 = op.inputs[0]
        input1 = op.inputs[1]
        input1_config = op.input_quant_config[1]
        input1_num_of_bits = input1_config.num_of_bits
        input1_n_size = input1.shape[-1]
        if input0.shape is None or input1.shape is None:
            logger.error("input shape is None")
            return op

        if input1_num_of_bits != 8 and input1_num_of_bits != 16:
            logger.warning(f"The num_of_bits of input1 {input1_num_of_bits} is not supported.")
            return op

        input0_dims = len(input0.shape)
        input1_dims = len(input1.shape)
        input1_orig_shape = input1.shape
        align = 16 if input1_num_of_bits == 8 else 8

        if op.platform in [
            TargetPlatform.ESPDL_C_INT8,
            TargetPlatform.ESPDL_C_INT16,
            TargetPlatform.ESPDL_C_H_PRE_INT16,
        ]:
            if input1_dims >= 2:
                # *CN -> *NC
                # Because esp-dl's Conv2D implementation in C requires the filter layout to be NHWC.
                insert_transpose_node(
                    graph=graph, var=op.inputs[1], op=op, perm=list(range(len(input1.shape) - 2)) + [-1, -2]
                )
                insert_reshape_node(graph=graph, var=op.inputs[1], op=op, shape=input1_orig_shape)
        else:
            # align
            if input1_n_size % align == 0:
                if input1_dims >= 2:
                    # *CN -> *(N/align)C(align) = *(N/align)HWC(align)
                    c, n = input1.shape[-2:]
                    insert_reshape_node(
                        graph=graph, var=op.inputs[1], op=op, shape=input1.shape[:-2] + [c, n // align, align]
                    )
                    insert_transpose_node(
                        graph=graph, var=op.inputs[1], op=op, perm=list(range(len(input1.shape) - 2)) + [-2, -3, -1]
                    )
                    insert_reshape_node(graph=graph, var=op.inputs[1], op=op, shape=input1_orig_shape)
            # unalign
            else:
                aligned_len = input1_n_size // align * align

                if input1_dims >= 2:
                    c, n = input1.shape[-2:]
                    trans_op = insert_transpose_node(
                        graph=graph, var=op.inputs[1], op=op, perm=list(range(len(input1.shape) - 2)) + [-1, -2]
                    )
                    if aligned_len > 0:
                        insert_slice_node(
                            graph=graph,
                            var=op.inputs[1],
                            op=op,
                            starts=[0] * len(input1.shape),
                            ends=input1.shape[:-2] + [aligned_len, c],
                            axes=list(range(len(input1.shape))),
                            steps=[1] * len(input1.shape),
                        )
                        insert_reshape_node(
                            graph=graph, var=op.inputs[1], op=op, shape=input1.shape[:-2] + [n // align, align, c]
                        )
                        insert_transpose_node(
                            graph=graph, var=op.inputs[1], op=op, perm=list(range(len(input1.shape) - 2)) + [-3, -1, -2]
                        )
                        insert_reshape_node(
                            graph=graph, var=op.inputs[1], op=op, shape=input1.shape[:-2] + [aligned_len, c]
                        )
                        # concat align and unalign
                        concat_op = insert_concat_node(
                            graph=graph,
                            insert_op_var=op.inputs[1],
                            insert_op=op,
                            link_vars=[trans_op.outputs[0]],
                            link_vars_src_op=[trans_op],
                            axis=-2,
                        )
                        # insert unalign
                        insert_slice_node(
                            graph=graph,
                            var=concat_op.inputs[1],
                            op=concat_op,
                            starts=[0] * (len(input1.shape) - 2) + [aligned_len, 0],
                            ends=input1.shape[:-2] + [input1_n_size, c],
                            axes=list(range(len(input1.shape))),
                            steps=[1] * len(input1.shape),
                        )
                        concat_axis = concat_op.attributes["axis"]
                        concat_op.outputs[0].shape[concat_axis] = 0
                        for input in concat_op.inputs:
                            concat_op.outputs[0].shape[concat_axis] = (
                                concat_op.outputs[0].shape[concat_axis] + input.shape[concat_axis]
                            )

                    insert_reshape_node(graph=graph, var=op.inputs[1], op=op, shape=input1_orig_shape)

        return op


class FuseReluLikePattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if not isinstance(op, QuantableOperation):
            return op

        # The FUSE_OP_PATTERNS may remove some ops.
        if op.name not in graph.operations:
            return op

        if op.type in ["Conv", "Gemm", "MatMul"]:
            op.attributes["activation"] = "Linear"
            downstream_op = graph.get_downstream_operations(op)
            if len(downstream_op) == 1:  # the downstream op have only one op and this op is relu
                # if downstream_op[0].type in ["Relu", "Clip"]:
                if downstream_op[0].type in ["Relu"]:
                    logger.debug(f"fuse {op.type}:{op.name} and {downstream_op[0].type}:{downstream_op[0].name}")
                    conv_quant_config = op.config
                    relu_quant_config = downstream_op[0].config
                    new_config = OperationQuantizationConfig(
                        conv_quant_config.input_quantization_config,
                        relu_quant_config.output_quantization_config,
                    )

                    # graph.remove_operation(downstream_op[0], keep_coherence=True)
                    graph = fuse_downstream_operation(graph, downstream_op[0], keep_coherence=True)
                    op.config = new_config
                    op.attributes["activation"] = downstream_op[0].type

        return op


class QuantVariableToIntPattern(OperationExporter):
    @staticmethod
    def calculate_exponent(config: TensorQuantizationConfig):
        if not config.policy.has_property(QuantizationProperty.LINEAR):
            raise ValueError("Critical Quantization Error! Non-linear config detected.")
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            raise ValueError("Critical Quantization Error! Asymmetrical config detected.")

        if not config.scale:
            return None

        exponent = None
        if config.policy.has_property(QuantizationProperty.PER_TENSOR) and config.policy.has_property(
            QuantizationProperty.POWER_OF_2
        ):
            scale = convert_any_to_numpy(config.scale)
            exponent = [int(np.log2(scale))]
        elif config.policy.has_property(QuantizationProperty.PER_CHANNEL) and config.policy.has_property(
            QuantizationProperty.POWER_OF_2
        ):
            scale = convert_any_to_numpy(config.scale)
            exponent = np.log2(scale).astype(int).tolist()
        return exponent

    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if not isinstance(op, QuantableOperation):
            logger.info("skip not QuantableOperation")
            return op

        # collect quantable vars, where we need to quantize parameters
        info = ExporterPatternInfo()

        for config, var in [_ for _ in op.config_with_variable]:
            if not var or not config:
                logger.info("skip not config or var")
                continue

            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not info.get_var_config(var.name):
                info.add_var_config(var.name, config)

            if not info.get_var_exponents(var.name):
                exponent = QuantVariableToIntPattern.calculate_exponent(config)
                if exponent:
                    info.add_var_exponents(var.name, exponent)
                    info.add_var_config(var.name, config)
                    logger.debug(f"{var.name} exponent: {exponent}")
                else:
                    logger.info(
                        "Skip %s from (op name:%s, type:%s) because it's not quantized" % (var.name, op.name, op.type)
                    )
            else:
                continue

            if var.is_parameter:
                assert len(var.dest_ops) == 1, (
                    f"Can not export variable {var.name}, cause it has more than 1 destination operations. "
                    "PPQ require all parameters to have only 1 destination operation."
                )

                # override quantization state, so that we can export parameter correctly.
                if config.state == QuantizationStates.BAKED:
                    config.state = QuantizationStates.ACTIVATED
                if config.state == QuantizationStates.PASSIVE_BAKED:
                    config.state = QuantizationStates.PASSIVE

                if config.policy.has_property(QuantizationProperty.LINEAR):
                    if (
                        op.platform
                        in [
                            TargetPlatform.ESPDL_H_PRE_INT16,
                            TargetPlatform.ESPDL_S3_H_PRE_INT16,
                            TargetPlatform.ESPDL_C_H_PRE_INT16,
                        ]
                    ) and config.num_of_bits >= 16:
                        var.value = PPQLinearQuant_toInt(tensor=var.value.type(dtype=torch.float64), config=config)
                    else:
                        var.value = PPQLinearQuant_toInt(tensor=var.value, config=config)
            elif not var.is_parameter:
                if config.policy.has_property(QuantizationProperty.LINEAR):
                    if config.num_of_bits == 8 and config.exponent_bits == 0:
                        var.dtype = DataType.INT8
                    elif config.num_of_bits == 16 and config.exponent_bits == 0:
                        var.dtype = DataType.INT16
                    else:
                        var.dtype = DataType.FP32

        return op


class ResetParamLayoutPattern(OperationExporter):
    def reset_conv_filter_layout(self, tensor, quant_type, group=None, platform=TargetPlatform.ESPDL_INT8):
        layout = LayoutAnnotation.NCHW

        if len(tensor.shape) != 3 and len(tensor.shape) != 4:
            logger.error(f"Conv filter don't support {len(tensor.shape)}D tensor.")
            return tensor, layout

        if len(tensor.shape) == 3:
            n, c, w = tensor.shape  # n denotes output channels, c denotes input channels,
            tensor = tensor.permute(0, 2, 1)  # NCW -> NWC

            if platform in [
                TargetPlatform.ESPDL_C_INT8,
                TargetPlatform.ESPDL_C_INT16,
                TargetPlatform.ESPDL_C_H_PRE_INT16,
            ]:
                layout = LayoutAnnotation.NWC
                if group > 1:
                    tensor = tensor.permute(1, 2, 0)  # depthwise: NWC -> WCN
                    layout = LayoutAnnotation.WCN
            else:
                align = 16 if quant_type == EspQuantType.S8 else 8
                aligned_len = n // align * align
                aligned_tensor = tensor[0:aligned_len, ...]
                aligned_tensor = aligned_tensor.reshape(n // align, align, w, c)  # NWC -> (N/align,align)WC
                # (N/align,align)WC -> (N/align)WC(align)
                aligned_tensor = aligned_tensor.permute(0, 2, 3, 1)
                # (N/align)WC(align) -> (aligned_len)WC
                aligned_tensor = aligned_tensor.reshape(aligned_len, w, c)

                if n % align != 0:
                    unaligned_tensor = tensor[aligned_len:n, ...]  # NWC
                    if group == 1 or group == None:
                        aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)
                    else:
                        n_remain = n - aligned_len
                        unaligned_tensor = unaligned_tensor.permute(2, 1, 0)  # depthwise unaligned: NWC -> CWN
                        unaligned_tensor = unaligned_tensor.reshape(n_remain, w, c)
                        aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)

                    if align == 16:
                        layout = LayoutAnnotation.N16WC16_UNALIGNED
                    else:
                        layout = LayoutAnnotation.N8WC8_UNALIGNED
                else:
                    if align == 16:
                        layout = LayoutAnnotation.N16WC16
                    else:
                        layout = LayoutAnnotation.N8WC8

                tensor = aligned_tensor

            # TODO:: modify the layout of depthwise conv in ESP-DL, keep it same with conv
            if group == 1 or group == None:
                tensor = tensor.reshape(w, c, n)  # reshape to WCN
            else:
                tensor = tensor.reshape(w, n, c)  # reshape to WNC

        elif len(tensor.shape) == 4:
            n, c, h, w = tensor.shape  # n denotes output channels, c denotes input channels,
            tensor = tensor.permute(0, 2, 3, 1)  # NCHW -> NHWC

            if platform in [
                TargetPlatform.ESPDL_C_INT8,
                TargetPlatform.ESPDL_C_INT16,
                TargetPlatform.ESPDL_C_H_PRE_INT16,
            ]:
                layout = LayoutAnnotation.NHWC
                if group is not None and group > 1:
                    tensor = tensor.permute(1, 2, 3, 0)  # depthwise: NHWC -> HWCN
                    layout = LayoutAnnotation.HWCN
            else:
                align = 16 if quant_type == EspQuantType.S8 else 8
                aligned_len = n // align * align
                aligned_tensor = tensor[0:aligned_len, ...]
                aligned_tensor = aligned_tensor.reshape(n // align, align, h, w, c)  # NHWC -> (N/align,align)HWC
                # (N/align,align)HWC -> (N/align)HWC(align)
                aligned_tensor = aligned_tensor.permute(0, 2, 3, 4, 1)
                # (N/align)HWC(align) -> (aligned_len)HWC
                aligned_tensor = aligned_tensor.reshape(aligned_len, h, w, c)

                if n % align != 0:
                    unaligned_tensor = tensor[aligned_len:n, ...]  # NHWC
                    if group == 1 or group == None:
                        aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)
                    else:
                        n_remain = n - aligned_len
                        unaligned_tensor = unaligned_tensor.permute(3, 1, 2, 0)  # depthwise unaligned: NHWC -> CHWN
                        unaligned_tensor = unaligned_tensor.reshape(n_remain, h, w, c)
                        aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)

                    if align == 16:
                        layout = LayoutAnnotation.N16HWC16_UNALIGNED
                    else:
                        layout = LayoutAnnotation.N8HWC8_UNALIGNED
                else:
                    if align == 16:
                        layout = LayoutAnnotation.N16HWC16
                    else:
                        layout = LayoutAnnotation.N8HWC8

                tensor = aligned_tensor

            # TODO:: modify the layout of depthwise conv in ESP-DL, keep it same with conv
            if group == 1 or group == None:
                tensor = tensor.reshape(h, w, c, n)  # reshape to HWCN
            else:
                tensor = tensor.reshape(h, w, n, c)  # reshape to HWNC

        return tensor, layout

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        quant_type = op.attributes.get("quant_type", None)
        if quant_type == None or quant_type == EspQuantType.F32 or not isinstance(op, QuantableOperation):
            return op

        info = ExporterPatternInfo()

        if op.type == "Conv":
            for var in op.inputs:
                if not var.is_parameter:
                    continue

                tensor = var.value
                if len(tensor.shape) == 3 or len(tensor.shape) == 4:  # Conv1d/Conv2d Filter
                    group = op.attributes.get("group", None)
                    aligned_tensor, layout = self.reset_conv_filter_layout(tensor, quant_type, group, op.platform)
                    info.add_var_layout(var.name, layout)
                    var.value = aligned_tensor
                    logger.debug(f"reset {op.type}:{op.name}, shape:{tensor.shape}, layout to {layout}")

        elif op.type == "Gemm":
            for var in op.inputs:
                if not var.is_parameter:
                    continue

                # fix the transB attribute is 0
                tensor = var.value
                alpha = op.attributes.get("alpha", 1.0)
                beta = op.attributes.get("beta", 1.0)
                assert alpha == 1.0 and beta == 1.0, "alpha and beta must be 1.0 and 0.0"

                if len(tensor.shape) == 2:  # Gemm Filter
                    trans_filter = op.attributes.get("transB", 0)
                    if trans_filter != 0:
                        logger.debug("transB is not 0, transpose the filter and reset transB")
                        op.attributes["transB"] = 0  # update 'transB'
                        tensor = tensor.transpose(1, 0)  # [N, C] -> [C, N]
                    tensor = tensor.unsqueeze(-1).unsqueeze(-1)  # CN -> CNHW
                    # CNHW -> NCHW, same with conv2d filter
                    tensor = tensor.permute(1, 0, 2, 3)

                    aligned_tensor, layout = self.reset_conv_filter_layout(tensor, quant_type, None, op.platform)
                    info.add_var_layout(var.name, layout)
                    var.value = aligned_tensor
                    logger.debug(f"reset {op.type}:{op.name}, shape:{var.value.shape}, layout to {layout}")

        elif op.type == "MatMul" and op.inputs[1].is_parameter and len(op.inputs[1].shape) >= 2:
            tensor = op.inputs[1].value
            tensor_orig_shape = tensor.shape
            c, n = tensor_orig_shape[-2:]
            tensor = tensor.reshape(-1, c, n)
            tensor_reset = []
            layout = None

            for tensor_tmp in tensor:
                tensor_tmp = tensor_tmp.unsqueeze(-1).unsqueeze(-1)  # CN -> CNHW
                # CNHW -> NCHW, same with conv2d filter
                tensor_tmp = tensor_tmp.permute(1, 0, 2, 3)

                tensor_tmp, layout = self.reset_conv_filter_layout(tensor_tmp, quant_type, None, op.platform)
                # HWCN -> CN
                tensor_tmp.squeeze(0).squeeze(0)
                tensor_reset.append(tensor_tmp)

            info.add_var_layout(op.inputs[1].name, layout)
            op.inputs[1].value = torch.stack(tensor_reset).reshape(tensor_orig_shape)

        return op


class AddLUTPattern(OperationExporter):
    def __init__(self, int16_step=1) -> None:
        super().__init__()
        self.int16_step = int(int16_step)  # the step of int16 LUT

    def get_scale(self, var: Variable, info: ExporterPatternInfo) -> torch.Tensor:
        exponent = info.get_var_exponents(var.name)
        scale = 1.0
        if exponent:
            if isinstance(exponent, list):
                scale = 2 ** exponent[0]
            else:
                scale = 2**exponent

        return scale

    def calculate_lut(
        self,
        op: QuantableOperation,
        info: ExporterPatternInfo,
        max: int,
        min: int,
        step: int = 1,
    ) -> torch.Tensor:
        # Get forward function
        platform_dispatching_table = OPERATION_FORWARD_TABLE[op.platform]
        if op.type not in platform_dispatching_table:
            raise NotImplementedError(
                f"Graph op: {op.name}({op.type}) "
                f"has no backend implementation on target platform {op.platform}. "
                "Register this op to esp_ppq.executor.base.py and esp_ppq.executor.op first"
            )
        operation_forward_func = platform_dispatching_table[op.type]

        # Calculate output and lut
        input = torch.arange(min, max + step, step=step, dtype=torch.float)
        input = input * self.get_scale(op.inputs[0], info)
        inputs = [input]

        if len(op.inputs) > 1:
            for op_input in op.inputs[1:]:
                inputs.append(op_input.value * self.get_scale(op_input, info))
        output = operation_forward_func(op, inputs)
        device = op.output_quant_config[0].scale.device
        lut = PPQLinearQuant_toInt(output.to(device), op.output_quant_config[0])

        return lut

    def get_lut_name(self, op: Operation, info: ExporterPatternInfo):
        index = len(info.luts)
        name = f"{op.type}_lut_{index}"
        return name

    def check_op(self, op: Operation):
        """
        True if this op can be implemented by LUT, otherwise False
        """

        if op.type == "PRelu":
            if op.inputs[1].value.numel() == 1:
                return True
            else:
                return False
        elif op.type == "Clip":
            return True
        elif len(op.outputs) > 1 or len(op.inputs) > 1:
            return False
        elif op.type in ACTIVATION_OP_SET or op.type in MATH_OP_SET:
            return True

        return False

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        quant_type = op.attributes.get("quant_type", None)
        if quant_type == None or quant_type == EspQuantType.F32 or not isinstance(op, QuantableOperation):
            return op

        info = ExporterPatternInfo()

        if self.check_op(op):
            lut = None
            if quant_type == EspQuantType.S8:
                lut = self.calculate_lut(op, info, 127, -128, 1)
            elif quant_type == EspQuantType.S16 and self.int16_step > 0:
                lut = self.calculate_lut(op, info, 2**15 - 1, -(2**15), self.int16_step)

            if lut != None:
                lut_name = self.get_lut_name(op, info)
                op.attributes["lut"] = lut_name
                info.add_lut(lut_name, lut, info.get_var_exponents(op.outputs[0].name))

        return op


class QuantRNNPattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ["GRU"] and isinstance(op, QuantableOperation):
            op.attributes["gate_exponent"] = GRU_QUANT_EXPONENT
        elif op.type in ["LSTM"] and isinstance(op, QuantableOperation):
            op.attributes["gate_exponent"] = LSTM_QUANT_EXPONENT

        return op



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TensorTypeAndShape(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TensorTypeAndShape()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensorTypeAndShape(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # TensorTypeAndShape
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TensorTypeAndShape
    def ElemType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TensorTypeAndShape
    def Shape(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.TensorShape import TensorShape
            obj = TensorShape()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def TensorTypeAndShapeStart(builder):
    builder.StartObject(2)

def Start(builder):
    TensorTypeAndShapeStart(builder)

def TensorTypeAndShapeAddElemType(builder, elemType):
    builder.PrependInt32Slot(0, elemType, 0)

def AddElemType(builder, elemType):
    TensorTypeAndShapeAddElemType(builder, elemType)

def TensorTypeAndShapeAddShape(builder, shape):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)

def AddShape(builder, shape):
    TensorTypeAndShapeAddShape(builder, shape)

def TensorTypeAndShapeEnd(builder):
    return builder.EndObject()

def End(builder):
    return TensorTypeAndShapeEnd(builder)

import FlatBuffers.Dl.TensorShape
try:
    from typing import Optional
except:
    pass

class TensorTypeAndShapeT(object):

    # TensorTypeAndShapeT
    def __init__(self):
        self.elemType = 0  # type: int
        self.shape = None  # type: Optional[FlatBuffers.Dl.TensorShape.TensorShapeT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        tensorTypeAndShape = TensorTypeAndShape()
        tensorTypeAndShape.Init(buf, pos)
        return cls.InitFromObj(tensorTypeAndShape)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, tensorTypeAndShape):
        x = TensorTypeAndShapeT()
        x._UnPack(tensorTypeAndShape)
        return x

    # TensorTypeAndShapeT
    def _UnPack(self, tensorTypeAndShape):
        if tensorTypeAndShape is None:
            return
        self.elemType = tensorTypeAndShape.ElemType()
        if tensorTypeAndShape.Shape() is not None:
            self.shape = FlatBuffers.Dl.TensorShape.TensorShapeT.InitFromObj(tensorTypeAndShape.Shape())

    # TensorTypeAndShapeT
    def Pack(self, builder):
        if self.shape is not None:
            shape = self.shape.Pack(builder)
        TensorTypeAndShapeStart(builder)
        TensorTypeAndShapeAddElemType(builder, self.elemType)
        if self.shape is not None:
            TensorTypeAndShapeAddShape(builder, shape)
        tensorTypeAndShape = TensorTypeAndShapeEnd(builder)
        return tensorTypeAndShape



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ValueInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ValueInfo()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsValueInfo(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # ValueInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ValueInfo
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ValueInfo
    def ValueInfoType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.TypeInfo import TypeInfo
            obj = TypeInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ValueInfo
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ValueInfo
    def Exponents(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # ValueInfo
    def ExponentsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # ValueInfo
    def ExponentsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ValueInfo
    def ExponentsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

def ValueInfoStart(builder):
    builder.StartObject(4)

def Start(builder):
    ValueInfoStart(builder)

def ValueInfoAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    ValueInfoAddName(builder, name)

def ValueInfoAddValueInfoType(builder, valueInfoType):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(valueInfoType), 0)

def AddValueInfoType(builder, valueInfoType):
    ValueInfoAddValueInfoType(builder, valueInfoType)

def ValueInfoAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    ValueInfoAddDocString(builder, docString)

def ValueInfoAddExponents(builder, exponents):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(exponents), 0)

def AddExponents(builder, exponents):
    ValueInfoAddExponents(builder, exponents)

def ValueInfoStartExponentsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartExponentsVector(builder, numElems):
    return ValueInfoStartExponentsVector(builder, numElems)

def ValueInfoEnd(builder):
    return builder.EndObject()

def End(builder):
    return ValueInfoEnd(builder)

import FlatBuffers.Dl.TypeInfo
try:
    from typing import List, Optional
except:
    pass

class ValueInfoT(object):

    # ValueInfoT
    def __init__(self):
        self.name = None  # type: str
        self.valueInfoType = None  # type: Optional[FlatBuffers.Dl.TypeInfo.TypeInfoT]
        self.docString = None  # type: str
        self.exponents = None  # type: List[int]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        valueInfo = ValueInfo()
        valueInfo.Init(buf, pos)
        return cls.InitFromObj(valueInfo)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, valueInfo):
        x = ValueInfoT()
        x._UnPack(valueInfo)
        return x

    # ValueInfoT
    def _UnPack(self, valueInfo):
        if valueInfo is None:
            return
        self.name = valueInfo.Name()
        if valueInfo.ValueInfoType() is not None:
            self.valueInfoType = FlatBuffers.Dl.TypeInfo.TypeInfoT.InitFromObj(valueInfo.ValueInfoType())
        self.docString = valueInfo.DocString()
        if not valueInfo.ExponentsIsNone():
            if np is None:
                self.exponents = []
                for i in range(valueInfo.ExponentsLength()):
                    self.exponents.append(valueInfo.Exponents(i))
            else:
                self.exponents = valueInfo.ExponentsAsNumpy()

    # ValueInfoT
    def Pack(self, builder):
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.valueInfoType is not None:
            valueInfoType = self.valueInfoType.Pack(builder)
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.exponents is not None:
            if np is not None and type(self.exponents) is np.ndarray:
                exponents = builder.CreateNumpyVector(self.exponents)
            else:
                ValueInfoStartExponentsVector(builder, len(self.exponents))
                for i in reversed(range(len(self.exponents))):
                    builder.PrependInt64(self.exponents[i])
                exponents = builder.EndVector()
        ValueInfoStart(builder)
        if self.name is not None:
            ValueInfoAddName(builder, name)
        if self.valueInfoType is not None:
            ValueInfoAddValueInfoType(builder, valueInfoType)
        if self.docString is not None:
            ValueInfoAddDocString(builder, docString)
        if self.exponents is not None:
            ValueInfoAddExponents(builder, exponents)
        valueInfo = ValueInfoEnd(builder)
        return valueInfo



from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Iterable, Union

import torch

from esp_ppq.api.setting import *
from esp_ppq.core import (
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
    TensorQuantizationConfig,
    empty_ppq_cache,
)
from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR import BaseGraph, GraphReplacer, Operation, QuantableGraph, QuantableOperation, QuantableVariable
from esp_ppq.IR.base.command import QuantizeOperationCommand
from esp_ppq.quantization.optim import *


class BaseQuantizer(metaclass=ABCMeta):
    def __init__(self, graph: BaseGraph, verbose: bool = True) -> None:
        if not isinstance(graph, BaseGraph):
            raise TypeError(
                f'To initialize a Quantizer, a BaseGraph instance is needed.'
                f' While {type(graph)} was givne, if your graph is maintained by GraphCommandProcessor, '
                'use GraphCommandProcessor.graph here instead.'
            )
        self._verbose = verbose
        self._graph = graph
        self._processor = QuantableGraph(GraphReplacer(self._graph))

    @empty_ppq_cache
    def quantize(
        self,
        inputs: Union[torch.Tensor, list, dict],
        calib_dataloader: Iterable,
        executor: BaseGraphExecutor,
        setting: QuantizationSetting,
        **kwargs,
    ) -> None:
        # step - 1, prequant pipeline:
        # prequant pipeline will change your network structure and float value.
        prequant_pipeline = self.build_prequant_pipeline(setting, executor=executor)
        prequant_pipeline.optimize(
            graph=self._graph, dataloader=calib_dataloader, executor=executor, verbose=self._verbose, **kwargs
        )

        # step - 2, quantize all operations
        executor.load_graph(self._graph)
        executor.tracing_operation_meta(inputs=inputs)

        for op_name, operation in self._graph.operations.items():
            if operation.platform == TargetPlatform.UNSPECIFIED:
                if operation.type in self.quant_operation_types:
                    operation.platform = self.target_platform
                else:
                    operation.platform = TargetPlatform.FP32

            if operation.platform not in {TargetPlatform.FP32, TargetPlatform.SOI}:
                self.quantize_operation(op_name)

        # quantize operation will modify network structure
        # it is necessary calling self._executor before further execution
        # step - 3, calling graph optimization pipeline
        executor.load_graph(self._graph)
        quant_pipeline = self.build_quant_pipeline(setting)

        quant_pipeline.optimize(
            graph=self._graph, dataloader=calib_dataloader, executor=executor, verbose=self._verbose, **kwargs
        )

        if self._verbose:
            print(self.report(), end='')
            print('Network Quantization Finished.')

    def quantize_operation(self, op_name: str, platform: TargetPlatform = None) -> QuantableOperation:
        if op_name not in self._graph.operations:
            raise KeyError(f'Can not find op {op_name} in your graph, chech operation name again.')
        converting_operation = self._graph.operations[op_name]
        if isinstance(converting_operation, QuantableOperation):
            ppq_warning(f'Operation {op_name} has been quantized, can not to quantize it twice.')
            return converting_operation

        # override platform with calling parameter.
        if platform is not None:
            converting_operation.platform = platform
        else:
            platform = converting_operation.platform

        if platform in {TargetPlatform.FP32, TargetPlatform.SOI}:
            return self._graph.operations[op_name]

        # if platform == TargetPlatform.UNSPECIFIED we can skip its quantization when type is not supported.
        if platform == TargetPlatform.UNSPECIFIED and converting_operation.type not in self.quant_operation_types:
            return self._graph.operations[op_name]

        # create quantize config and convert operation.
        self._processor(
            QuantizeOperationCommand(
                op_name=op_name,
                target_platform=platform,
                config=self.init_quantize_config(operation=converting_operation),
            )
        )
        return self._graph.operations[op_name]

    @staticmethod
    def create_default_quant_config(
        op: Operation,
        num_of_bits: int = 8,
        quant_min: Union[int, float] = -127,
        quant_max: Union[int, float] = 128,
        observer_algorithm: str = 'percentile',
        policy: QuantizationPolicy = QuantizationPolicy(
            QuantizationProperty.PER_TENSOR + QuantizationProperty.LINEAR + QuantizationProperty.SYMMETRICAL
        ),
        rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
        exponent_bits: int = 0,
    ) -> OperationQuantizationConfig:
        """
        为你的算子创建一个默认量化信息

        对于一个 Onnx 算子而言，它总是会有几个输入和输出 Variable
        你需要为每一个相连的 Variable 初始化量化信息 TensorQuantConfig

        这个函数就是用来帮你初始化这些信息的。

        一个麻烦的问题是：

            对于很多 onnx 算子而言，他们的部分输入都是不需要量化的:

            如 Clip 算子的三个输入 value, min, max, 大部分框架不要求量化 min, max
            如 Reshape 算子的两个输入 value, shape, 其中 shape 不能够被量化

            PPQ 的算子接线器中记录了这些信息

            算子接线器中记录了所有标准 onnx 的默认量化策略
            该函数将使用预定义的算子量化策略初始化量化信息

        你可以在 Quantizer 中对默认量化策略进行进一步修改

        Create a default quantization configuration for given op.
        For each onnx op, there will be some input and output variables.

        You are required to create tensor quantization config for every
        input and output variables.

        This function is designed for creating a default quantization config for you.

            The created OQC(Op Quantization Config) is based on OpSocket.

            In fact, there are some rules or templates when creating the OQC:
            For Clip Op which has 3 input variable, namely value, min and max
                most framework does not require a quantization config for min and max.
            For Reshape Op which has 2 input variable, namely value and shape
                the input shape can never be quantized.

        Those rules are pre-defined within OpSocket, thus ppq will create
        OQC based on underlying OpSocket of your Op.

        After the default OQC got created, you can overwrite its state in quantizer.
        """
        assert isinstance(op, Operation), (
            f'Can only initialize OQC for esp_ppq.IR.Operation, however {type(op)} was given.'
        )
        assert isinstance(policy, QuantizationPolicy), (
            'Can not create quantization config - Quantization Policy Type Error.'
        )
        assert isinstance(rounding, RoundingPolicy), 'Can not create quantization config - Rounding Policy Type Error.'

        socket = op.socket
        input_cfgs, output_cfgs = [], []
        for index in range(op.num_of_input):
            state = QuantizationStates.INITIAL
            # for those unexpected inputs and outputs
            # ppq just initilize them as normal variable.
            if index < len(socket.in_plat):
                target_plat = socket.in_plat[index]
                if target_plat == TargetPlatform.FP32:
                    state = QuantizationStates.FP32
                if target_plat == TargetPlatform.SOI:
                    state = QuantizationStates.FP32
            input_cfgs.append(
                TensorQuantizationConfig(
                    policy=policy,
                    rounding=rounding,
                    num_of_bits=num_of_bits,
                    scale=None,
                    offset=None,
                    exponent_bits=exponent_bits,
                    quant_min=quant_min,
                    quant_max=quant_max,
                    observer_algorithm=observer_algorithm,
                    state=state,
                )
            )

        for index in range(op.num_of_output):
            state = QuantizationStates.INITIAL
            # for those unexpected inputs and outputs
            # ppq just initilize them as normal variable.
            if index < len(socket.out_plat):
                target_plat = socket.out_plat[index]
                if target_plat == TargetPlatform.FP32:
                    state = QuantizationStates.FP32
                if target_plat == TargetPlatform.SOI:
                    state = QuantizationStates.FP32
            output_cfgs.append(
                TensorQuantizationConfig(
                    policy=policy,
                    rounding=rounding,
                    num_of_bits=num_of_bits,
                    scale=None,
                    offset=None,
                    exponent_bits=exponent_bits,
                    quant_min=quant_min,
                    quant_max=quant_max,
                    observer_algorithm=observer_algorithm,
                    state=state,
                )
            )

        return OperationQuantizationConfig(input_cfgs, output_cfgs)

    @abstractmethod
    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        raise NotImplementedError('Implement this first.')

    @abstractproperty
    @property
    def quant_operation_types(self) -> set:
        raise NotImplementedError('Quantizier does not have a quantable op set yet.')

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.INT8

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}

    def report(self) -> str:
        debug_str = ''
        # stats:
        quant_ops = [op for op in self._graph.operations.values() if isinstance(op, QuantableOperation)]
        quant_vars = [var for var in self._graph.variables.values() if isinstance(var, QuantableVariable)]
        quant_cfgs = []

        config_states_cnt = dict.fromkeys(QuantizationStates, 0)
        for op in quant_ops:
            for cfg, _ in op.config_with_variable:
                config_states_cnt[cfg.state] += 1
                quant_cfgs.append(cfg)

        debug_str += '--------- Network Snapshot ---------\n'
        debug_str += f'Num of Op:                    [{len(self._graph.operations)}]\n'
        debug_str += f'Num of Quantized Op:          [{len(quant_ops)}]\n'
        debug_str += f'Num of Variable:              [{len(self._graph.variables)}]\n'
        debug_str += f'Num of Quantized Var:         [{len(quant_vars)}]\n'
        debug_str += '------- Quantization Snapshot ------\n'
        debug_str += f'Num of Quant Config:          [{len(quant_cfgs)}]\n'
        for state, cnt in config_states_cnt.items():
            if cnt <= 0:
                continue
            padding_str = ' ' * max(28 - len(state.name), 0)
            debug_str += f'{state.name}:{padding_str} [{cnt}]\n'
        return debug_str

    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        assert isinstance(setting, QuantizationSetting), (
            f'PPQ needs a OptimSetting instance to initialize optimization pipeline, however {type(setting)} was given.'
        )

        if setting.matrix_factorization == True:
            ppq_warning(
                'PPQ Matrix Factorization Pass has been removed from QuantizationSetting since 0.6.5, this pass must be called manually now.'
            )
            ppq_warning(
                'PPQ Matrix Factorization Pass 已经不能通过 QuantizationSetting 调用，现在你必须手动调用该优化过程'
            )

        list_of_passes = []
        if setting.ssd_equalization:
            equalization_setting = setting.ssd_setting
            list_of_passes.append(
                SSDEqualizationPass(
                    optimize_level=equalization_setting.opt_level,
                    channel_ratio=equalization_setting.channel_ratio,
                    loss_threshold=equalization_setting.loss_threshold,
                    layer_norm=equalization_setting.layer_norm,
                    iteration=equalization_setting.iteration,
                )
            )

        if setting.fusion:
            fusion_setting = setting.fusion_setting
            list_of_passes.append(
                QuantizeFusionPass(
                    fuse_activation=fusion_setting.fuse_activation,
                    fuse_passive_op=fusion_setting.fuse_passive_op,
                    activation_type=self.activation_fusion_types,
                )
            )

            if fusion_setting.remove_useless_quantization:
                list_of_passes.append(QuantizeSimplifyPass())

        if setting.quantize_parameter:
            param_setting = setting.quantize_parameter_setting
            list_of_passes.append(ParameterQuantizePass(method=param_setting.calib_algorithm))

        if setting.quantize_activation:
            act_setting = setting.quantize_activation_setting
            list_of_passes.append(RuntimeCalibrationPass(method=act_setting.calib_algorithm))

            if setting.quant_config_modify:
                quant_config_modify_setting = setting.quant_config_modify_setting
                list_of_passes.append(
                    QuantConfigModifyPass(
                        custom_config=quant_config_modify_setting.custom_config,
                    )
                )

        if setting.fusion:
            if fusion_setting.align_quantization:
                list_of_passes.append(
                    QuantAlignmentPass(
                        elementwise_alignment=fusion_setting.align_elementwise_to,
                        concat_alignment=fusion_setting.align_concat_to,
                        pooling_alignment=fusion_setting.align_avgpooling_to,
                        resize_alignment=fusion_setting.align_resize_to,
                        force_overlap=fusion_setting.force_alignment_overlap,
                    )
                )

        if setting.quantize_parameter:
            param_setting = setting.quantize_parameter_setting
            if param_setting.quantize_passive_parameter:
                list_of_passes.append(
                    PassiveParameterQuantizePass(
                        clip_visiblity=param_setting.clip_visiblity, pad_visiblity=param_setting.pad_visiblity
                    )
                )

        if setting.bias_correct:
            bias_correct_setting = setting.bias_correct_setting
            list_of_passes.append(
                BiasCorrectionPass(
                    block_size=bias_correct_setting.block_size,
                    interested_layers=bias_correct_setting.interested_layers,
                    steps=bias_correct_setting.steps,
                    collecting_device=bias_correct_setting.collecting_device,
                )
            )

        if setting.lsq_optimization:
            lsq_setting = setting.lsq_optimization_setting
            list_of_passes.append(
                LearnedStepSizePass(
                    interested_layers=lsq_setting.interested_layers,
                    lr=lsq_setting.lr,
                    collecting_device=lsq_setting.collecting_device,
                    steps=lsq_setting.steps,
                    gamma=lsq_setting.gamma,
                    is_scale_trainable=lsq_setting.is_scale_trainable,
                    block_size=lsq_setting.block_size,
                )
            )
            # requant passive parameters
            param_setting = setting.quantize_parameter_setting
            list_of_passes.append(
                PassiveParameterQuantizePass(
                    clip_visiblity=param_setting.clip_visiblity, pad_visiblity=param_setting.pad_visiblity
                )
            )

        if setting.blockwise_reconstruction:
            blockwise_reconstruction_setting = setting.blockwise_reconstruction_setting
            list_of_passes.append(
                AdaroundPass(
                    interested_layers=blockwise_reconstruction_setting.interested_layers,
                    lr=blockwise_reconstruction_setting.lr,
                    collecting_device=blockwise_reconstruction_setting.collecting_device,
                    steps=blockwise_reconstruction_setting.steps,
                    gamma=blockwise_reconstruction_setting.gamma,
                    is_scale_trainable=blockwise_reconstruction_setting.is_scale_trainable,
                    block_size=blockwise_reconstruction_setting.block_size,
                )
            )
            # requant passive parameters
            param_setting = setting.quantize_parameter_setting
            list_of_passes.append(
                PassiveParameterQuantizePass(
                    clip_visiblity=param_setting.clip_visiblity, pad_visiblity=param_setting.pad_visiblity
                )
            )

        if setting.quantize_parameter:
            if param_setting.baking_parameter:
                list_of_passes.append(ParameterBakingPass())

        if setting.extension:
            list_of_passes.append(ExtensionPass(setting.extension_setting.my_first_parameter))

        return QuantizationOptimizationPipeline(passes=list_of_passes)

    def build_prequant_pipeline(
        self, setting: QuantizationSetting, executor: BaseGraphExecutor
    ) -> QuantizationOptimizationPipeline:
        assert isinstance(setting, QuantizationSetting), (
            f'PPQ needs a OptimSetting instance to initialize optimization pipeline, however {type(setting)} was given.'
        )

        list_of_passes = []
        if setting.weight_split:
            weight_split_setting = setting.weight_split_setting
            list_of_passes.append(
                HorizontalLayerSplitPass(
                    interested_layers=weight_split_setting.interested_layers,
                    method=weight_split_setting.method,
                    value_threshold=weight_split_setting.value_threshold,
                )
            )

        if setting.channel_split:
            channel_split_setting = setting.channel_split_setting
            list_of_passes.append(
                ChannelwiseSplitPass(
                    optimize_level=channel_split_setting.opt_level,
                    iterations=channel_split_setting.iterations,
                    threshold=channel_split_setting.value_threshold,
                    including_bias=channel_split_setting.including_bias,
                    including_act=channel_split_setting.including_act,
                    bias_multiplier=channel_split_setting.bias_multiplier,
                    act_multiplier=channel_split_setting.act_multiplier,
                    interested_layers=channel_split_setting.interested_layers,
                )
            )

        if setting.equalization:
            equalization_setting = setting.equalization_setting
            list_of_passes.append(
                LayerwiseEqualizationPass(
                    optimize_level=equalization_setting.opt_level,
                    iterations=equalization_setting.iterations,
                    value_threshold=equalization_setting.value_threshold,
                    including_bias=equalization_setting.including_bias,
                    including_act=equalization_setting.including_act,
                    bias_multiplier=equalization_setting.bias_multiplier,
                    act_multiplier=equalization_setting.act_multiplier,
                    interested_layers=equalization_setting.interested_layers,
                )
            )

        if setting.convtranspose_decomposition:
            list_of_passes.append(ConvTransposeDecompositionPass())

        return QuantizationOptimizationPipeline(passes=list_of_passes)



import time
from typing import Callable

import onnxruntime
import pandas as pd
import torch
import torch.utils.data
import torchvision.datasets as datasets
import torchvision.transforms as transforms
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Subset
from tqdm import tqdm

from esp_ppq.executor.torch import TorchExecutor
from esp_ppq.IR import BaseGraph


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k
    prec1, prec5 = accuracy(output.data, target, topk=(1, 5))
    """
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def load_imagenet_from_directory(
    directory: str,
    subset: int = None,
    batchsize: int = 32,
    shuffle: bool = False,
    require_label: bool = True,
    num_of_workers: int = 12,
) -> torch.utils.data.DataLoader:
    """
    一套十分标准的 Imagenet 数据加载流程，
    directory: 数据加载的位置
    subset: 如果设置为非空，则从数据集中抽取subset大小的子集
    require_label: 是否需要标签
    shuffle: 是否打乱数据集
    """
    dataset = datasets.ImageFolder(
        directory,
        transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        ),
    )

    if subset:
        dataset = Subset(dataset, indices=[_ for _ in range(0, subset)])
    if require_label:
        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=batchsize,
            shuffle=shuffle,
            num_workers=num_of_workers,
            pin_memory=False,
            drop_last=True,  # onnx 模型不支持动态 batchsize，最后一个批次的数据尺寸可能不对齐，因此丢掉最后一个批次的数据
        )
    else:
        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=batchsize,
            shuffle=shuffle,
            num_workers=num_of_workers,
            pin_memory=False,
            collate_fn=lambda x: torch.cat([sample[0].unsqueeze(0) for sample in x], dim=0),
            drop_last=False,  # 不需要标签的数据为 calib 数据，无需 drop
        )


def evaluate_torch_module_with_imagenet(
    model: torch.nn.Module,
    imagenet_validation_dir: str = None,
    batchsize: int = 32,
    device: str = 'cuda',
    imagenet_validation_loader: DataLoader = None,
    verbose: bool = True,
) -> pd.DataFrame:
    model.eval()
    with torch.no_grad():
        model_forward_function = lambda input_tensor: model(input_tensor)
        return _evaluate_any_module_with_imagenet(
            model_forward_function=model_forward_function,
            batchsize=batchsize,
            device=device,
            imagenet_validation_dir=imagenet_validation_dir,
            imagenet_validation_loader=imagenet_validation_loader,
            verbose=verbose,
        )


def evaluate_onnx_module_with_imagenet(
    onnxruntime_model_path: str,
    imagenet_validation_dir: str = None,
    batchsize: int = 32,
    device: str = 'cuda',
    imagenet_validation_loader: DataLoader = None,
    verbose: bool = True,
) -> pd.DataFrame:
    sess = onnxruntime.InferenceSession(path_or_bytes=onnxruntime_model_path, providers=['CUDAExecutionProvider'])
    input_placeholder_name = sess.get_inputs()[0].name
    with torch.no_grad():
        # TODO 可能导致一些内存资源浪费
        model_forward_function = lambda input_tensor: torch.tensor(
            sess.run(input_feed={input_placeholder_name: input_tensor.cpu().numpy()}, output_names=None)
        )[0]
        return _evaluate_any_module_with_imagenet(
            model_forward_function=model_forward_function,
            batchsize=batchsize,
            device=device,
            imagenet_validation_dir=imagenet_validation_dir,
            imagenet_validation_loader=imagenet_validation_loader,
            verbose=verbose,
        )


def evaluate_mmlab_module_with_imagenet(
    model: torch.nn.Module,
    imagenet_validation_dir: str = None,
    batchsize: int = 32,
    device: str = 'cuda',
    imagenet_validation_loader: DataLoader = None,
    verbose: bool = True,
) -> pd.DataFrame:
    model.eval()
    with torch.no_grad():
        model_forward_function = lambda input_tensor: model.forward_test(input_tensor, img_metas={})
        return _evaluate_any_module_with_imagenet(
            model_forward_function=model_forward_function,
            batchsize=batchsize,
            device=device,
            imagenet_validation_dir=imagenet_validation_dir,
            imagenet_validation_loader=imagenet_validation_loader,
            verbose=verbose,
        )


def evaluate_ppq_module_with_imagenet(
    model: BaseGraph,
    imagenet_validation_dir: str = None,
    batchsize: int = 32,
    device: str = 'cuda',
    imagenet_validation_loader: DataLoader = None,
    verbose: bool = True,
) -> pd.DataFrame:
    """
    一套用来测试 ppq 模块的逻辑，
    直接送入 esp_ppq.IR.BaseGraph 就好了
    """

    executor = TorchExecutor(graph=model, device=device)
    model_forward_function = lambda input_tensor: torch.tensor(executor(*[input_tensor])[0])
    return _evaluate_any_module_with_imagenet(
        model_forward_function=model_forward_function,
        batchsize=batchsize,
        device=device,
        imagenet_validation_dir=imagenet_validation_dir,
        imagenet_validation_loader=imagenet_validation_loader,
        verbose=verbose,
    )


def _evaluate_any_module_with_imagenet(
    model_forward_function: Callable,
    imagenet_validation_dir: str,
    batchsize: int = 32,
    device: str = 'cuda',
    imagenet_validation_loader: DataLoader = None,
    verbose: bool = True,
):
    """
    一套十分标准的imagenet测试逻辑
    """

    recorder = {'loss': [], 'top1_accuracy': [], 'top5_accuracy': [], 'batch_time': []}

    if imagenet_validation_loader is None:
        imagenet_validation_loader = load_imagenet_from_directory(
            imagenet_validation_dir, batchsize=batchsize, shuffle=False
        )

    loss_fn = torch.nn.CrossEntropyLoss().to('cpu')

    for batch_idx, (batch_input, batch_label) in tqdm(
        enumerate(imagenet_validation_loader), desc='Evaluating Model...', total=len(imagenet_validation_loader)
    ):
        batch_input = batch_input.to(device)
        batch_label = batch_label.to(device)
        batch_time_mark_point = time.time()

        batch_pred = model_forward_function(batch_input)
        if isinstance(batch_pred, list):
            batch_pred = torch.tensor(batch_pred)

        recorder['batch_time'].append(time.time() - batch_time_mark_point)
        recorder['loss'].append(loss_fn(batch_pred.to('cpu'), batch_label.to('cpu')))
        prec1, prec5 = accuracy(torch.tensor(batch_pred).to('cpu'), batch_label.to('cpu'), topk=(1, 5))
        recorder['top1_accuracy'].append(prec1.item())
        recorder['top5_accuracy'].append(prec5.item())

        if batch_idx % 100 == 0 and verbose:
            print(
                'Test: [{0} / {1}]\tPrec@1 {top1:.3f} ({top1:.3f})\tPrec@5 {top5:.3f} ({top5:.3f})'.format(
                    batch_idx,
                    len(imagenet_validation_loader),
                    top1=sum(recorder['top1_accuracy']) / len(recorder['top1_accuracy']),
                    top5=sum(recorder['top5_accuracy']) / len(recorder['top5_accuracy']),
                )
            )

    if verbose:
        print(
            ' * Prec@1 {top1:.3f} Prec@5 {top5:.3f}'.format(
                top1=sum(recorder['top1_accuracy']) / len(recorder['top1_accuracy']),
                top5=sum(recorder['top5_accuracy']) / len(recorder['top5_accuracy']),
            )
        )

    # dump records toward dataframe
    dataframe = pd.DataFrame()
    for column_name in recorder:
        dataframe[column_name] = recorder[column_name]
    return dataframe



import argparse
import os
import struct
import sys

import numpy as np
import tensorrt as trt

INPUT_H = 32
INPUT_W = 32
CHANNEL = 1
OUTPUT_SIZE = 10
INPUT_BLOB_NAME = "input.1"
OUTPUT_BLOB_NAME = "32"
TRT_LOGGER = trt.Logger(trt.Logger.INFO)
EXPLICIT_BATCH = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)


def GiB(val):
    return val * 1 << 30


def generateLenetEngine(weight_map, json_file, engine_path):
    builder = trt.Builder(TRT_LOGGER)
    network = builder.create_network(EXPLICIT_BATCH)
    config = builder.create_builder_config()

    # Adds an input to the network.
    data = network.add_input(INPUT_BLOB_NAME, trt.float32, (1, CHANNEL, INPUT_H, INPUT_W))
    assert data
    data.name = "input.1"

    conv1 = network.add_convolution(
        input=data,
        num_output_maps=6,
        kernel_shape=(5, 5),
        kernel=weight_map["conv1.weight"],
        bias=weight_map["conv1.bias"],
    )

    assert conv1
    conv1.stride = (1, 1)
    conv1.get_output(0).name = "onnx::Relu_11"

    relu1 = network.add_activation(conv1.get_output(0), type=trt.ActivationType.RELU)
    assert relu1
    relu1.get_output(0).name = "onnx::Pad_12"

    pool1 = network.add_pooling(input=relu1.get_output(0), window_size=trt.DimsHW(2, 2), type=trt.PoolingType.AVERAGE)
    assert pool1
    pool1.stride = (2, 2)
    pool1.get_output(0).name = "onnx::AveragePool_14"

    conv2 = network.add_convolution(
        pool1.get_output(0), 16, trt.DimsHW(5, 5), weight_map["conv2.weight"], weight_map["conv2.bias"]
    )

    assert conv2
    conv2.stride = (1, 1)
    conv2.get_output(0).name = "input"

    relu2 = network.add_activation(conv2.get_output(0), type=trt.ActivationType.RELU)
    assert relu2
    relu2.get_output(0).name = "onnx::Relu_16"

    pool2 = network.add_pooling(input=relu2.get_output(0), window_size=trt.DimsHW(2, 2), type=trt.PoolingType.AVERAGE)
    assert pool2
    pool2.stride = (2, 2)
    pool2.get_output(0).name = "onnx::Pad_17"

    fc1 = network.add_fully_connected(
        input=pool2.get_output(0), num_outputs=120, kernel=weight_map['fc1.weight'], bias=weight_map['fc1.bias']
    )
    assert fc1
    fc1.get_output(0).name = "onnx::AveragePool_19"

    relu3 = network.add_activation(fc1.get_output(0), type=trt.ActivationType.RELU)
    assert relu3
    relu3.get_output(0).name = "onnx::Relu_27"

    fc2 = network.add_fully_connected(
        input=relu3.get_output(0), num_outputs=84, kernel=weight_map['fc2.weight'], bias=weight_map['fc2.bias']
    )
    assert fc2
    fc2.get_output(0).name = "onnx::Gemm_28"

    relu4 = network.add_activation(fc2.get_output(0), type=trt.ActivationType.RELU)
    assert relu4
    relu4.get_output(0).name = "onnx::Relu_29"

    fc3 = network.add_fully_connected(
        input=relu4.get_output(0), num_outputs=OUTPUT_SIZE, kernel=weight_map['fc3.weight'], bias=weight_map['fc3.bias']
    )
    assert fc3
    fc3.get_output(0).name = "onnx::Gemm_30"

    prob = network.add_softmax(fc3.get_output(0))
    assert prob
    prob.get_output(0).name = "32"

    network.mark_output(prob.get_output(0))
    config.max_workspace_size = GiB(1)

    config.set_flag(trt.BuilderFlag.INT8)
    setDynamicRange(network, json_file)

    engine = builder.build_engine(network, config)

    with open(engine_path, "wb") as f:
        f.write(engine.serialize())

    del weight_map
    del engine
    del network
    del builder


def setDynamicRange(network, json_file):
    """Sets ranges for network layers."""
    import json

    with open(json_file) as json_file_info:
        quant_param_json = json.load(json_file_info)

    act_quant = quant_param_json["act_quant_info"]
    for i in range(network.num_inputs):
        input_tensor = network.get_input(i)
        if act_quant.__contains__(input_tensor.name):
            print("\033[1;32mWrite quantization parameters:%s\033[0m" % input_tensor.name)
            value = act_quant[input_tensor.name]
            tensor_max = abs(value)
            tensor_min = -abs(value)
            input_tensor.dynamic_range = (tensor_min, tensor_max)

    for i in range(network.num_layers):
        layer = network.get_layer(i)

        for output_index in range(layer.num_outputs):
            tensor = layer.get_output(output_index)
            if act_quant.__contains__(tensor.name):
                print("\033[1;32mWrite quantization parameters:%s\033[0m" % tensor.name)
                value = act_quant[tensor.name]
                tensor_max = abs(value)
                tensor_min = -abs(value)
                tensor.dynamic_range = (tensor_min, tensor_max)
            else:
                print("\033[1;31mNo quantization parameters are written: %s\033[0m" % tensor.name)


def load_weights(file):
    print(f"Loading weights: {file}")

    assert os.path.exists(file), 'Unable to load weight file.'

    weight_map = {}
    with open(file, "r") as f:
        lines = [line.strip() for line in f]
    count = int(lines[0])
    assert count == len(lines) - 1
    for i in range(1, count + 1):
        splits = lines[i].split(" ")
        name = splits[0]
        cur_count = int(splits[1])
        assert cur_count + 2 == len(splits)
        values = []
        for j in range(2, len(splits)):
            # hex string to bytes to float
            values.append(struct.unpack(">f", bytes.fromhex(splits[j])))
        weight_map[name] = np.array(values, dtype=np.float32)
    return weight_map


if __name__ == '__main__':
    sys.argv[1]  # Weight parameter
    sys.argv[2]  # Quantization parameter json
    sys.argv[3]  # tensorrt engine file
    weight_map = load_weights(sys.argv[1])

    generateLenetEngine(weight_map, sys.argv[2], sys.argv[3])



from typing import Iterable, List, Set

import torch

from esp_ppq.core import (
    ALIGNMENT_MANUL_OVERRIDE,
    PASSIVE_OPERATIONS,
    TYPES_FOR_ALIGNMENT,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
    TensorQuantizationConfig,
    empty_ppq_cache,
    ppq_warning,
)
from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation, Variable
from esp_ppq.IR.quantize import QuantableVariable
from esp_ppq.IR.search import SearchableGraph
from esp_ppq.quantization.observer.range import minmax_to_scale_offset

from .base import QuantizationOptimizationPass


class QuantizeSimplifyPass(QuantizationOptimizationPass):
    """
    ## PPQ Quantize Simplify Pass(通用量化精简过程)

    PPQ use Tensor Quantization Configuration(A data structure defined in esp_ppq.core) to
    control quantization. Each quantable op will have a list of TQC as its quantization config,
    which contains necessary quantization parameter(scale, offset), in order to quantize its input(s) and output(s).

    While TQC is a powerful tool for describing quantization, it introduces some undiserible features:

    For a subgraph like:

            Relu1 - Relu2

    PPQ will create at least 4 TQC here, namely the input TQC of Relu1 and Relu2, and the output TQC of Relu1 and Relu2.
    Problem here is the output TQC of Relu1 and the input TQC of Relu2 is actually duplicated, the output variable
    should not be quantized twice.

    This Simplify Pass will detect all the duplicated TQCs in your network, disable them and create a link with their
    dominating TQCs. Disabled TQC will have and inactive state(QuantizationState.OVERRLAPED), so PPQ executor will
    simply ignore them when executing.

    A duplicated TQC is an input TQC(A) whose binding variable has been quantized by another output TQC(B),
    and the input TQC(A) should have the same bit-width as the output TQC(B)

    ### Parameters:

    * No Parameter

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True
        setting.fusion_setting.remove_useless_quantization = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(self) -> None:
        super().__init__(name='PPQ Quantize Simplify Pass')

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for _, variable in graph.variables.items():
            assert isinstance(variable, Variable)
            source_op = variable.source_op

            if source_op is None:
                continue  # input variables in network, they do not have a source
            if not isinstance(source_op, QuantableOperation):
                continue
            source_config = source_op.config.output_quantization_config[source_op.outputs.index(variable)]

            if source_config.state in {QuantizationStates.FP32}:
                continue  # if source config does not have a valid state, skip it.

            for downstream_op, dest_idx in zip(variable.dest_ops, variable.dest_idx):
                if downstream_op is None:
                    continue  # output variables in network, they do not have a destination
                if not isinstance(downstream_op, QuantableOperation):
                    continue

                input_config = downstream_op.config.input_quantization_config[dest_idx]
                if source_op.platform == downstream_op.platform:
                    if input_config.state == QuantizationStates.INITIAL and input_config.is_same_scheme(source_config):
                        input_config.dominated_by = source_config


class QuantizeFusionPass(QuantizationOptimizationPass):
    """
    ## PPQ Quantize Fusion Pass(通用量化图融合过程)

    Operation fusion (or kernel/layer fusion) is key optimization in many state-of-the-art execution frameworks.

    Graph fusion can combine operations into a single op to obtain higher accuracy and performance,
        Pattern like: Conv + Relu can be reduced to ConvRelu. This fusion will reduce memory accesses,
        and the quantization point after conv can also be removed.

    Technically we can fuse those layers before quantization, while fused layers are not supported by onnx standard.
        So to say ConvRelu is not a valid onnx operation, no execution framework can parse it.

    Therefore, PPQ will simulate the graph fusion by adjusting quantization config: if PPQ finds their is a
        pattern like Conv + Relu, the output quantization of Conv will be disabled, pretending that the Conv + Relu
        fusion has happened.

    This Pass is designed for 2 types graph fusion:
        1. activation fusion
        2. passive operation fusion

    For activation fusion, PPQ will identify the pattern: Computing op + Activation Op from your network. The output
        quantization of computing op will be disabled with their state being set to QuantizationState.OVERLAPPED.

    Activation fusion here supports only simple activation patterns,
        for complex activation functions like mish, swish,
        will be represented as mish = tanh + mul + softplus, swish = sigmoid + mul in onnx,
        cause onnx does not have a op defination for them.
        Identifying those complex patterns requires pattern matching, which is implemented in esp_ppq.IR.search.py

    Complex quantization fusions must be invoked manually, PPQ implemented softplus & swish fusion functions in
        esp_ppq.quantization.optim.refine.MishFusionPass
        esp_ppq.quantization.optim.refine.SwishFusionPass

    For passive operation fusion, PPQ will keep the input and the output variable share a same scale for passive operations.
        An operation is identified as passive op only if its attribute "is_active_quant_op" = False, this
        attribute is initialized by quantizer.

    If there is a passive operation having multiple input and output, the fusion procedure will make its
    FIRST input variable and ALL output variables share the same scale(the same scale as its first input).
    The quantization states of all output variables will be set to QuantizationState.OVERLAPPED.

    ### Parameters:

    * activation_type(Set[str]):

            A collection contains all activation types.

            The pattern will be recognized as [Computing Op -> Activation Op],

            By graph fusion, the output quantization of the Computing Op and
                the input quantization of the activation op will be disabled.

    * fuse_activation(bool)

            Whether to fuse activation op with computing op.

    # fuse_passive_op(bool)

            Whether to fuse passive op so that the input variable and output variable will share a same scale.

    * fuse_matmul_add(bool)

            Fuse MatMul + Bias Add

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(
        self,
        activation_type: Set[str],
        fuse_activation: bool = True,
        fuse_passive_op: bool = True,
        fuse_relu_clip: bool = True,
    ) -> None:
        self.fuse_activation = fuse_activation
        self.fuse_passive_op = fuse_passive_op
        self.fuse_relu_clip = fuse_relu_clip
        self.activation_types = activation_type
        super().__init__(name='PPQ Quantization Fusion Pass')

    def is_same_platform(self, operations: List[Operation]):
        platforms = [operation.platform for operation in operations]
        return all([platform == platforms[0] for platform in platforms])

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        processor = SearchableGraph(graph)

        # fuse computing operations and its following activation.
        if self.fuse_activation:
            patterns = processor.pattern_matching(
                patterns=[lambda x: x.is_computing_op, lambda x: x.type in self.activation_types],
                edges=[[0, 1]],
                exclusive=True,
            )

            for computing_op, act_op in patterns:
                if not isinstance(act_op, QuantableOperation):
                    continue
                if not isinstance(computing_op, QuantableOperation):
                    continue

                if (
                    computing_op.platform != act_op.platform
                    and computing_op.config.output_quantization_config[0].state != QuantizationStates.FP32
                ):
                    ppq_warning(
                        f'Unexpected dispatching was found: '
                        f'Op {computing_op.name} and {act_op.name} should be send to a same platform.'
                    )
                    continue

                if (
                    len(graph.get_downstream_operations(computing_op)) == 1
                    and len(graph.get_upstream_operations(act_op)) == 1
                ):
                    computing_op.config.output_quantization_config[
                        0
                    ].dominated_by = act_op.config.output_quantization_config[0]
                    act_op.config.input_quantization_config[0].dominated_by = act_op.config.output_quantization_config[
                        0
                    ]

            if 'Swish' in self.activation_types:
                search_engine = SearchableGraph(graph)
                patterns = search_engine.pattern_matching(
                    patterns=[lambda x: x.is_computing_op, 'Sigmoid', 'Mul'],
                    edges=[[0, 1], [1, 2], [0, 2]],
                    exclusive=True,
                )

                for pattern in patterns:
                    if any([not isinstance(op, QuantableOperation) for op in pattern]):
                        ppq_warning(
                            f'There is a pattern of swish activation in your network start from {pattern[0]}, '
                            'however part of your swish activation is not quantable, '
                            'so that graph fusion can not merge their quantization configuration.'
                        )
                        continue
                    if any([op.platform != pattern[0].platform for op in pattern]):
                        ppq_warning(
                            f'There is a pattern of swish activation in your network start from {pattern[0]}, '
                            'however part of your swish activation is not quantable, '
                            'so that graph fusion can not merge their quantization configuration.'
                        )
                        continue
                    computing, sigmoid, mul = pattern

                    assert isinstance(computing, QuantableOperation)
                    assert isinstance(sigmoid, QuantableOperation)
                    assert isinstance(mul, QuantableOperation)

                    master_config = mul.config.output_quantization_config[0]
                    computing.config.output_quantization_config[0].dominated_by = master_config
                    sigmoid.config.input_quantization_config[0].dominated_by = master_config
                    sigmoid.config.output_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[1].dominated_by = master_config

            if 'Mish' in self.activation_types:
                search_engine = SearchableGraph(graph)
                patterns = search_engine.pattern_matching(
                    patterns=[lambda x: x.is_computing_op, 'Softplus', 'Tanh', 'Mul'],
                    edges=[[0, 1], [1, 2], [2, 3], [0, 3]],
                    exclusive=True,
                )

                for pattern in patterns:
                    if any([not isinstance(op, QuantableOperation) for op in pattern]):
                        ppq_warning(
                            f'There is a pattern of mish activation in your network start from {pattern[0]}, '
                            'however part of your mish activation is not quantable, '
                            'so that graph fusion can not merge their quantization configuration.'
                        )
                        continue
                    if any([op.platform != pattern[0].platform for op in pattern]):
                        ppq_warning(
                            f'There is a pattern of mish activation in your network start from {pattern[0]}, '
                            'however part of your mish activation is not quantable, '
                            'so that graph fusion can not merge their quantization configuration.'
                        )
                        continue
                    computing, softplus, tanh, mul = pattern

                    assert isinstance(computing, QuantableOperation)
                    assert isinstance(softplus, QuantableOperation)
                    assert isinstance(tanh, QuantableOperation)
                    assert isinstance(mul, QuantableOperation)

                    master_config = mul.config.output_quantization_config[0]
                    computing.config.output_quantization_config[0].dominated_by = master_config
                    tanh.config.input_quantization_config[0].dominated_by = master_config
                    tanh.config.output_quantization_config[0].dominated_by = master_config
                    softplus.config.input_quantization_config[0].dominated_by = master_config
                    softplus.config.output_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[1].dominated_by = master_config

        if self.fuse_passive_op:
            # all passive operations should never changes quantization configuration of its input
            # so to say their input and output share a same scale.
            for op in graph.operations.values():
                if op.type not in PASSIVE_OPERATIONS:
                    continue
                source_op = op.inputs[0].source_op
                if source_op is None:
                    continue  # beginning op, can not merge.
                if isinstance(op, QuantableOperation) and self.is_same_platform([op, source_op]):
                    TQC = op.config.input_quantization_config[0]
                    for output_cfg in op.config.output_quantization_config:
                        output_cfg.dominated_by = TQC

        if self.fuse_relu_clip:
            patterns = processor.pattern_matching(
                patterns=[lambda x: True, lambda x: x.type in {'Relu', 'Clip'}], edges=[[0, 1]], exclusive=True
            )
            for computing_op, act_op in patterns:
                if not isinstance(act_op, QuantableOperation):
                    continue
                if not isinstance(computing_op, QuantableOperation):
                    continue

                if (
                    len(graph.get_downstream_operations(computing_op)) == 1
                    and len(graph.get_upstream_operations(act_op)) == 1
                ):
                    computing_op.config.output_quantization_config[
                        0
                    ].dominated_by = act_op.config.output_quantization_config[0]
                    act_op.config.input_quantization_config[0].dominated_by = act_op.config.output_quantization_config[
                        0
                    ]


class QuantAlignmentPass(QuantizationOptimizationPass):
    """
    ## PPQ Quant Alignment Pass(通用量化对齐过程)

    When deploy on real hardware and inference framework,
        we will find that there are various restrictions or rules that we have to follow.

    * AVERAGE_POOL_2D: Input and outputs must all have same scale/zero_point

    * CONCATENATION: Input and outputs must all have same scale/zero_point

    * SLICE: Input and outputs must all have same scale/zero_point

    More detailed restrictions please refer to: https://www.tensorflow.org/lite/performance/quantization_spec

    Those restrictions, can be concluded as some quantization should share
        the same quantization parameter with others. PPQ Quant Alignment Pass is designed
        for dealing with problems like this.

    PPQ uses Tensor Quantization Config (A data structure defined in esp_ppq.core) to control the
        quantization logic, so to say if we want to align quantization parameters, we align
        their TQC in fact.

    The way to align TQC is simple, code like:
        tqc1.set_master(master=tqc2)
    Will make tqc1 and tqc2 share the same quantization parameters as tqc1 has, and change the
    state of tqc2 to be QuantizationState.PASSIVE

    If we access the scale of tqc2, PPQ will return its master TQC's scale instead, so does offset.

    That is tqc1 and tqc2 are bonuded with statement "tqc1.set_master(master=tqc2)".

    ### Parameters:

    * elementwise_alignment(Set[str]):

            Alignment method for elementwise ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All elementwise ops are listed in esp_ppq.core.common.py

    * concat_alignment(Set[str])

            Alignment method for concat-like ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All concat-like ops are listed in esp_ppq.core.common.py

    * averagepool_alignment(Set[str])

            Alignment method for pooling-like ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All pooling-like ops are listed in esp_ppq.core.common.py

    * resize_alignment(Set[str])

            Alignment method for Resize op.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

    * force_overlap(bool)

            TQC alignment might cause serious cascade effect.

            For subgraph like this:

            Conv1 ---
                    + --- Add1
            Conv2 ---
                    + --- Conv3

            If we demand Add1 to have same input scale, this alignment will affect Conv3 also,
                cause Conv2's output is feed to both Add1 and Conv3.

            If force_overlap = False, PPQ alignment procedure will remain the output scale of
                Conv2 as unchanged, while only align the input scale of Add1.

            If force_overlap = True, the input of Add1, Conv3 and the output of Conv2 will all
                be aligned to a same scale.

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True
        setting.fusion_setting.force_alignment_overlap = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(
        self,
        elementwise_alignment: str = 'Align to Large',
        concat_alignment: str = 'Align to Output',
        pooling_alignment: str = 'None',
        resize_alignment: str = 'None',
        logical_elementwise_alignment: str = 'Align to Large',
        force_overlap: bool = False,
        rnn_alignment: str = 'RNN Align to First',
    ) -> None:
        self.pooling_alignment = pooling_alignment
        self.elementwise_alignment = elementwise_alignment
        self.concat_alignment = concat_alignment
        self.resize_alignment = resize_alignment
        self.logical_elementwise_alignment = logical_elementwise_alignment
        self.force_overlap = force_overlap
        self.rnn_alignment = rnn_alignment
        assert self.elementwise_alignment in {'Align to Large', 'Align to Output', 'None'}, (
            'Elementwise Alignment can only be (None), (Align to Large) or (Align to Output)'
        )
        assert self.concat_alignment in {'Align to Large', 'Align to Output', 'None'}, (
            'Concat Alignment can only be (None), (Align to Large) or (Align to Output)'
        )
        assert self.pooling_alignment in {'Align to Output', 'None', 'Align to Input'}, (
            'Alignment method can only be (None), (Align to Input) or (Align to Output)'
        )
        assert self.resize_alignment in {'Align to Output', 'None', 'Align to Input'}, (
            'concat_merge_method can only be (None), (Align to Input) or (Align to Output)'
        )
        assert self.logical_elementwise_alignment in {'Align to Large', 'None'}, (
            'Logical Elementwise Alignment can only be (None) or (Align to Large)'
        )
        assert self.rnn_alignment in {'RNN Align to First', 'None'}, (
            'GRU/LSTM Alignment can only be (None) or (RNN Align to First)'
        )
        super().__init__(name='PPQ Quantization Alignment Pass')

    def align_to_input(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to input config. All output configs
        would share a same scale and offset with output config. (as a slave to
        input config)

        Any change to slave config will be rejected since then.
        """
        master_config = op.config.input_quantization_config[0]
        for slave_config in op.config.output_quantization_config:
            if slave_config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            if slave_config.state not in {QuantizationStates.FP32, QuantizationStates.SOI}:
                slave_config.master_by = master_config
        return master_config

    def align_to_first_output(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to input config. All output configs
        would share a same scale and offset with output config. (as a slave to
        input config)

        Any change to slave config will be rejected since then.
        """
        master_config = op.config.output_quantization_config[0]
        slave_configs = [op.config.output_quantization_config[1], op.config.input_quantization_config[5]]

        for slave_config in slave_configs:
            if slave_config.policy.has_property(QuantizationProperty.FLOATING):
                return master_config

            if slave_config.state not in {QuantizationStates.FP32, QuantizationStates.SOI}:
                slave_config.master_by = master_config
        return master_config

    def align_to_large(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to larger input config. The first input
        config will be set as master config, all slave config will share the
        same scale and offset with master.

        Any change to slave config will be rejected since then.
        """
        global_min, global_max, master_config = 0, 0, op.config.input_quantization_config[0]
        for config in op.config.input_quantization_config:
            if config.state in {QuantizationStates.FP32, QuantizationStates.SOI}:
                continue
            if config.policy.has_property(QuantizationProperty.FLOATING):
                continue

            assert config.policy.has_property(QuantizationProperty.PER_TENSOR), (
                'Quant Alignment can only happen with per tensor quantization.'
            )
            local_min = config.scale * (config.quant_min - config.offset)
            local_max = config.scale * (config.quant_max - config.offset)

            assert isinstance(local_min, torch.Tensor)
            assert isinstance(local_max, torch.Tensor)
            global_max = max(global_max, local_max.item())
            global_min = min(global_min, local_min.item())

        # recompute scale and offset
        scale, offset = minmax_to_scale_offset(global_min, global_max, op.config.input_quantization_config[0])

        device = master_config.scale.device
        master_config._dominator = master_config
        master_config.state = QuantizationStates.PASSIVE
        master_config.scale = torch.tensor(scale, dtype=torch.float32, device=device)
        master_config.offset = torch.tensor(offset, dtype=torch.float32, device=device)

        for slave_config in op.config.input_quantization_config[1:]:
            if config.state in {QuantizationStates.FP32, QuantizationStates.SOI}:
                continue
            if config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            slave_config.master_by = master_config
        return master_config

    def align_to_output(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to output config. All input configs
        would share a same scale and offset with output config. (as a slave to
        output config)

        Any change to slave config will be rejected since then.
        """
        master_config = op.config.output_quantization_config[0]
        for slave_config in op.config.input_quantization_config:
            if slave_config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            if slave_config.state not in {QuantizationStates.FP32, QuantizationStates.SOI}:
                slave_config.master_by = master_config
        return master_config

    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation):
                continue

            master_config = None
            if operation.type in TYPES_FOR_ALIGNMENT['Elementwise']:
                if self.elementwise_alignment == 'None':
                    continue
                if self.elementwise_alignment == 'Align to Large':
                    master_config = self.align_to_large(operation)
                else:
                    master_config = self.align_to_output(operation)

            elif operation.type in TYPES_FOR_ALIGNMENT['Concat']:
                if self.concat_alignment == 'None':
                    continue
                if self.concat_alignment == 'Align to Large':
                    master_config = self.align_to_large(operation)
                else:
                    master_config = self.align_to_output(operation)

            elif operation.type in TYPES_FOR_ALIGNMENT['Pooling']:
                if self.pooling_alignment == 'None':
                    continue
                if self.pooling_alignment == 'Align to Input':
                    self.align_to_input(operation)  # do not set master_config
                if self.pooling_alignment == 'Align to Output':
                    master_config = self.align_to_output(operation)
                if self.pooling_alignment == 'Align to Large':
                    raise ValueError('Alignment Method Error, Pooling Op can not align to lager input.')

            elif operation.type == 'Resize':
                if self.resize_alignment == 'None':
                    continue
                if self.resize_alignment == 'Align to Output':
                    master_config = self.align_to_output(operation)
                if self.resize_alignment == 'Align to Input':
                    self.align_to_input(operation)  # do not set master_config
                if self.resize_alignment == 'Align to Large':
                    raise ValueError('Alignment Method Error, Resize Op can not align to lager.')

            elif operation.type in TYPES_FOR_ALIGNMENT['LogicalElementwise']:
                if self.logical_elementwise_alignment == 'None':
                    continue
                if self.logical_elementwise_alignment == 'Align to Large':
                    master_config = self.align_to_large(operation)
            elif operation.type in TYPES_FOR_ALIGNMENT['RNN']:
                if self.rnn_alignment == 'None':
                    continue
                if self.rnn_alignment == 'RNN Align to First':
                    master_config = self.align_to_first_output(operation)
            elif ALIGNMENT_MANUL_OVERRIDE in operation.extension_attrib:
                method = operation.extension_attrib[ALIGNMENT_MANUL_OVERRIDE]
                if self.concat_alignment == 'Align to Large':
                    master_config = self.align_to_large(operation)
                elif self.concat_alignment == 'Align to Large':
                    master_config = self.align_to_output(operation)
                else:
                    ppq_warning(f'Unrecognized Alignment Method {method} for operation {operation.name}')

            if master_config is not None:
                # override up stream layer's config if possible
                for up_op in graph.get_upstream_operations(operation):
                    if not isinstance(up_op, QuantableOperation):
                        continue

                    if (
                        up_op.type in PASSIVE_OPERATIONS or len(graph.get_downstream_operations(up_op)) != 1
                    ) and not self.force_overlap:
                        continue
                    # for cfg, var in up_op.config_with_variable:
                    for cfg, var in zip(up_op.config.output_quantization_config, up_op.outputs):
                        if operation in var.dest_ops:
                            cfg.master_by = master_config


class SwishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__('Swish Fusion')

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda x: x.is_computing_op, 'Sigmoid', 'Mul'], edges=[[0, 1], [1, 2], [0, 2]], exclusive=True
        )

        for pattern in patterns:
            if any([not isinstance(op, QuantableOperation) for op in pattern]):
                ppq_warning(
                    f'There is a pattern of swish activation in your network start from {pattern[0]}, '
                    'however part of your swish activation is not quantable, '
                    'so that graph fusion can not merge their quantization configuration.'
                )
                continue
            if any([op.platform != pattern[0].platform for op in pattern]):
                ppq_warning(
                    f'There is a pattern of swish activation in your network start from {pattern[0]}, '
                    'however part of your swish activation is not quantable, '
                    'so that graph fusion can not merge their quantization configuration.'
                )
                continue
            computing, sigmoid, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(sigmoid, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            master_config = mul.config.output_quantization_config[0]
            computing.config.output_quantization_config[0].dominated_by = master_config
            sigmoid.config.input_quantization_config[0].dominated_by = master_config
            sigmoid.config.output_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[1].dominated_by = master_config


class MishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__('Mish Fusion')

    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda x: x.is_computing_op, 'Softplus', 'Tanh', 'Mul'],
            edges=[[0, 1], [1, 2], [2, 3], [0, 3]],
            exclusive=True,
        )

        for pattern in patterns:
            if any([not isinstance(op, QuantableOperation) for op in pattern]):
                ppq_warning(
                    f'There is a pattern of mish activation in your network start from {pattern[0]}, '
                    'however part of your mish activation is not quantable, '
                    'so that graph fusion can not merge their quantization configuration.'
                )
                continue
            if any([op.platform != pattern[0].platform for op in pattern]):
                ppq_warning(
                    f'There is a pattern of mish activation in your network start from {pattern[0]}, '
                    'however part of your mish activation is not quantable, '
                    'so that graph fusion can not merge their quantization configuration.'
                )
                continue
            computing, softplus, tanh, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(softplus, QuantableOperation)
            assert isinstance(tanh, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            master_config = mul.config.output_quantization_config[0]
            computing.config.output_quantization_config[0].dominated_by = master_config
            tanh.config.input_quantization_config[0].dominated_by = master_config
            tanh.config.output_quantization_config[0].dominated_by = master_config
            softplus.config.input_quantization_config[0].dominated_by = master_config
            softplus.config.output_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[1].dominated_by = master_config


class NxpInputRoundingRefinePass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name='PPQ Input Quantization Refine Pass')

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for variable in graph.variables.values():
            if isinstance(variable, QuantableVariable):
                if variable.source_op is None or not isinstance(variable.source_op, QuantableOperation):
                    for config in variable.dest_op_configs:
                        if config is None:
                            continue
                        config.rounding = RoundingPolicy.ROUND_HALF_DOWN


class NxpQuantizeFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name='PPQ Quantization Fusion Pass')

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        processor = SearchableGraph(graph)

        relu_fusion_matching = processor.activation_matching(start_op_types=['Conv', 'Add'], end_types=['Relu'])
        for conv_name, activation_names in relu_fusion_matching.items():
            conv = graph.operations[conv_name]
            if not isinstance(conv, QuantableOperation):
                continue
            if len(activation_names) == 1:
                activation = graph.operations[activation_names[0]]
                if not isinstance(activation, QuantableOperation):
                    continue
                activation_cfg = activation.config.output_quantization_config[0]
                conv_cfg = conv.config.output_quantization_config[0]
                conv_cfg.dominated_by = activation_cfg
                conv_cfg.state = QuantizationStates.OVERLAPPED

        concat_fusion_matching = processor.concat_matching(relay_pattern=lambda x, y: False, end_pattern=lambda _: True)
        for concat_name, upstream_layer_collection in concat_fusion_matching.items():
            concat = graph.operations[concat_name]
            if not isinstance(concat, QuantableOperation):
                continue
            for upstream_layer_name in upstream_layer_collection:
                upstream_layer = graph.operations[upstream_layer_name]
                if not isinstance(upstream_layer, QuantableOperation):
                    continue
                upstream_cfg = upstream_layer.config.output_quantization_config[0]
                concat_cfg = concat.config.output_quantization_config[0]
                upstream_cfg.dominated_by = concat_cfg
                upstream_cfg.state = QuantizationStates.OVERLAPPED



# This file is created by Nvidia Corp.
# Modified by PPQ develop team.
#
# Copyright 2020 NVIDIA Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import struct
from typing import List

from esp_ppq.core import NetworkFramework, QuantizationPolicy, QuantizationProperty, ppq_info, ppq_warning
from esp_ppq.IR import BaseGraph, GraphExporter
from esp_ppq.IR.quantize import QuantableOperation

from .onnxruntime_exporter import OnnxExporter, ONNXRUNTIMExporter


class TensorRTExporter_QDQ(ONNXRUNTIMExporter):
    """
    TensorRT PPQ 0.6.4 以来新加入的功能

    你需要注意，只有 TensorRT 8.0 以上的版本支持读取 PPQ 导出的量化模型
    并且 TensorRT 对于量化模型的解析存在一些 Bug，

    如果你遇到模型解析不对的问题，欢迎随时联系我们进行解决。

        已知的问题包括：
        1. 模型导出时最好不要包含其他 opset，如果模型上面带了别的opset，比如 mmdeploy，trt有可能会解析失败
        2. 模型导出时可能出现 Internal Error 10, Invalid Node xxx()，我们还不知道如何解决该问题

    Args:
        ONNXRUNTIMExporter (_type_): _description_
    """

    def export(self, file_path: str, graph: BaseGraph, save_as_external_data: bool = False) -> None:
        # step 1, export onnx file.
        super().export(
            file_path=file_path,
            graph=graph,
            quantized_param=False,
            config_path=None,
            save_as_external_data=save_as_external_data,
        )


class TensorRTExporter_JSON(GraphExporter):
    """
    使用该导出器导出符合 TensorRT 格式要求的量化模型。

    PPQ 可以将 TQC 中的量化信息导出到 json，并遵循 TensorRT 支持的格式。
    用户可以通过 Json 文件将图中所有 variable 的 min, max 传递给 TensorRT
    所有未写入 min, max 的变量将保持 fp32(fp16) 精度执行。
    """

    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        act_quant_info = {}
        for op in graph.topological_sort():
            if not isinstance(op, QuantableOperation):
                continue
            if op.type in {"Gather", "Unsqueeze", "Concat", "Reshape", "Squeeze"}:
                continue

            for cfg, var in op.config_with_variable:
                if not cfg.can_export(export_overlapped=True):
                    continue
                if var.is_parameter:
                    continue

                if cfg.policy != QuantizationPolicy(
                    QuantizationProperty.LINEAR + QuantizationProperty.SYMMETRICAL + QuantizationProperty.PER_TENSOR
                ):
                    ppq_warning(
                        f'Can not export quantization config on variable {var.name}, Quantization Policy is invalid.'
                    )
                    continue

                if cfg.num_of_bits != 8 or cfg.quant_max != 127 or cfg.quant_min != -128:
                    ppq_warning(
                        f'Can not export quantization config on variable {var.name}, '
                        'Tensor Quantization Config has unexpected setting.'
                    )
                    continue

                act_quant_info[var.name] = cfg.scale.item() * 127

        json_qparams_str = json.dumps({'act_quant_info': act_quant_info}, indent=4)
        with open(config_path, "w") as json_file:
            json_file.write(json_qparams_str)

    def export_weights(self, graph: BaseGraph, config_path: str = None):
        topo_order = graph.topological_sort()
        weights_list = []
        for index, op in enumerate(topo_order):
            if op.type in {"Conv", "Gemm"}:
                weights_list.extend(op.parameters)

        weight_file_path = os.path.join(os.path.dirname(config_path), "quantized.wts")

        f = open(weight_file_path, 'w')
        f.write("{}\n".format(len(weights_list)))

        for param in weights_list:
            weight_name = param.name
            weight_value = param.value.reshape(-1).cpu().numpy()
            f.write("{} {}".format(weight_name, len(weight_value)))
            for value in weight_value:
                f.write(" ")
                f.write(struct.pack(">f", float(value)).hex())
            f.write("\n")
        ppq_info(f'Parameters have been saved to file: {weight_file_path}')

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        input_shapes: List[List[int]] = [[1, 3, 224, 224]],
    ):
        ppq_info(
            'You are exporting PPQ Graph to TensorRT(Onnx + Json). \n'
            'Please Compile the TensorRT INT8 engine manually: \n\n'
            'from esp_ppq.utils.TensorRTUtil import build_engine \n'
            "build_engine(onnx_file='Quantized.onnx', int8_scale_file='Quantized.json', engine_file='Quantized.engine', int8=True)\n"
        )
        if config_path is not None:
            self.export_quantization_config(config_path, graph)
            self.export_weights(graph, config_path)
        _, ext = os.path.splitext(file_path)
        if ext == '.onnx' or graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



import os
import warnings

if os.path.dirname(os.path.realpath(__file__)) == os.path.join(os.path.realpath(os.getcwd()), "esp_ppq"):
    message = "You are importing esp_ppq within its own root folder ({}). "
    warnings.warn(message.format(os.getcwd()))

# This file defines export functions & class of esp_ppq.
from esp_ppq.api.setting import (
    ActivationQuantizationSetting,
    DispatchingTable,
    EqualizationSetting,
    GraphFormatSetting,
    LSQSetting,
    ParameterQuantizationSetting,
    QuantizationFusionSetting,
    QuantizationSetting,
    QuantizationSettingFactory,
    TemplateSetting,
)
from esp_ppq.core import *
from esp_ppq.executor import BaseGraphExecutor, TorchExecutor, TorchQuantizeDelegator
from esp_ppq.IR import (
    BaseGraph,
    GraphBuilder,
    GraphCommand,
    GraphExporter,
    GraphFormatter,
    Operation,
    QuantableGraph,
    SearchableGraph,
    TrainableGraph,
    Variable,
)
from esp_ppq.IR.deploy import RunnableGraph
from esp_ppq.IR.quantize import QuantableOperation, QuantableVariable
from esp_ppq.log import NaiveLogger
from esp_ppq.quantization.analyse import (
    graphwise_error_analyse,
    layerwise_error_analyse,
    parameter_analyse,
    statistical_analyse,
    variable_analyse,
)
from esp_ppq.quantization.measure import (
    torch_cosine_similarity,
    torch_cosine_similarity_as_loss,
    torch_KL_divergence,
    torch_mean_square_error,
    torch_snr_error,
)
from esp_ppq.quantization.optim import (
    BiasCorrectionPass,
    GRUSplitPass,
    HorizontalLayerSplitPass,
    LayerwiseEqualizationPass,
    MetaxGemmSplitPass,
    MishFusionPass,
    NxpInputRoundingRefinePass,
    NxpQuantizeFusionPass,
    NXPResizeModeChangePass,
    ParameterBakingPass,
    ParameterQuantizePass,
    PassiveParameterQuantizePass,
    QuantizationOptimizationPass,
    QuantizationOptimizationPipeline,
    QuantizeFusionPass,
    QuantizeSimplifyPass,
    RuntimeCalibrationPass,
    SwishFusionPass,
)
from esp_ppq.quantization.qfunction import (
    BaseQuantFunction,
    PPQDyamicLinearQuantFunction,
    PPQFloatingQuantFunction,
    PPQLinearQuant_toInt,
    PPQLinearQuantFunction,
    PPQuantFunction,
    PPQuantFunction_toInt,
)
from esp_ppq.quantization.quantizer import (
    BaseQuantizer,
    NXP_Quantizer,
    PPL_DSP_Quantizer,
    PPLCUDAQuantizer,
    TensorRTQuantizer,
)
from esp_ppq.scheduler import AggresiveDispatcher, ConservativeDispatcher, GraphDispatcher, PPLNNDispatcher
from esp_ppq.scheduler.perseus import Perseus
from esp_ppq.utils.round import ppq_numerical_round, ppq_round_to_power_of_2, ppq_tensor_round



import torch


def torch_KL_divergence(hist: torch.Tensor, ref_hist: torch.Tensor, eps=1e-30) -> float:
    if hist.ndim != 1 or ref_hist.ndim != 1:
        raise ValueError(
            'Only 1 dimension tensor can compute KL divergence with another tensor. '
            f'While your input hist has dimension {hist.ndim} and ref_hist has dimension {ref_hist.ndim}'
        )
    if len(hist) != len(ref_hist):
        raise ValueError('Can not compute KL divergence, len(hist) != len(ref_hist')

    # here we compute KL divergence at float64 precision, make sure your hist and ref_hist are stored at cpu.
    # otherwise it might be very slow.
    return torch.dot(hist.double(), torch.log10(hist.double() + eps) - torch.log10(ref_hist.double() + eps)).item()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class OperatorSetId(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = OperatorSetId()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsOperatorSetId(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # OperatorSetId
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OperatorSetId
    def Domain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # OperatorSetId
    def Version(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

def OperatorSetIdStart(builder):
    builder.StartObject(2)

def Start(builder):
    OperatorSetIdStart(builder)

def OperatorSetIdAddDomain(builder, domain):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(domain), 0)

def AddDomain(builder, domain):
    OperatorSetIdAddDomain(builder, domain)

def OperatorSetIdAddVersion(builder, version):
    builder.PrependInt64Slot(1, version, 0)

def AddVersion(builder, version):
    OperatorSetIdAddVersion(builder, version)

def OperatorSetIdEnd(builder):
    return builder.EndObject()

def End(builder):
    return OperatorSetIdEnd(builder)


class OperatorSetIdT(object):

    # OperatorSetIdT
    def __init__(self):
        self.domain = None  # type: str
        self.version = 0  # type: int

    @classmethod
    def InitFromBuf(cls, buf, pos):
        operatorSetId = OperatorSetId()
        operatorSetId.Init(buf, pos)
        return cls.InitFromObj(operatorSetId)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, operatorSetId):
        x = OperatorSetIdT()
        x._UnPack(operatorSetId)
        return x

    # OperatorSetIdT
    def _UnPack(self, operatorSetId):
        if operatorSetId is None:
            return
        self.domain = operatorSetId.Domain()
        self.version = operatorSetId.Version()

    # OperatorSetIdT
    def Pack(self, builder):
        if self.domain is not None:
            domain = builder.CreateString(self.domain)
        OperatorSetIdStart(builder)
        if self.domain is not None:
            OperatorSetIdAddDomain(builder, domain)
        OperatorSetIdAddVersion(builder, self.version)
        operatorSetId = OperatorSetIdEnd(builder)
        return operatorSetId



# Since ppq 0.6.4, PPQ_CONFIG.USING_CUDA_KERNEL = False is the defualt execution option in esp_ppq.
# However you should notice that if you are able to compile ppq kernel functions, the execution speed wiil boost at least 3x.
# This example will show you how to enable kernel function within esp_ppq.
# if you want to use kernel function everywhere, just rewrite esp_ppq.core.config.PPQ_CONFIG.USING_CUDA_KERNEL = True

from typing import Iterable

import torch
import torchvision
from torch.utils.data import DataLoader

from esp_ppq import BaseGraph, QuantizationSettingFactory, TargetPlatform
from esp_ppq.api import ENABLE_CUDA_KERNEL, export_ppq_graph, quantize_torch_model

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda'  # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.PPL_CUDA_INT8  # identify a target platform for your network.


def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


# Load a pretrained mobilenet v2 model
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# use this to wrap up your code
# all functions inside ENABLE_CUDA_KERNEL will switch to ppq kernel functions.
with ENABLE_CUDA_KERNEL():
    # create a setting for quantizing your network with PPL CUDA.
    quant_setting = QuantizationSettingFactory.pplcuda_setting()
    quant_setting.equalization = True  # use layerwise equalization algorithm.
    quant_setting.dispatcher = 'conservative'  # dispatch this network in conservertive way.
    quant_setting.lsq_optimization = True  # finetune your network.

    # Load training data for creating a calibration dataloader.
    calibration_dataset = load_calibration_dataset()
    calibration_dataloader = DataLoader(dataset=calibration_dataset, batch_size=BATCHSIZE, shuffle=True)

    # quantize your model.
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=calibration_dataloader,
        calib_steps=32,
        input_shape=[BATCHSIZE] + INPUT_SHAPE,
        setting=quant_setting,
        collate_fn=collate_fn,
        platform=PLATFORM,
        onnx_export_file='Output/onnx.model',
        device=DEVICE,
        verbose=0,
    )

    # Quantization Result is a PPQ BaseGraph instance.
    assert isinstance(quantized, BaseGraph)

    # export quantized graph.
    export_ppq_graph(
        graph=quantized,
        platform=PLATFORM,
        graph_save_to='Output/quantized(onnx).onnx',
        config_save_to='Output/quantized(onnx).json',
    )



# ------------------------------------------------------------
# PPQ 最佳实践示例工程，在这个工程中，我们将向你展示如何充分调动 PPQ 的各项功能
# ------------------------------------------------------------
import torch
import torchvision

from esp_ppq import *
from esp_ppq.api import *

# ------------------------------------------------------------
# Step - 1. 加载校准集与模型
# ------------------------------------------------------------
BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8
CALIBRATION = [torch.rand(size=INPUT_SHAPE) for _ in range(32)]
QS = QuantizationSettingFactory.default_setting()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# Step - 2. 执行首次量化，完成逐层误差分析
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        setting=QS,
        onnx_export_file='Output/onnx.model',
        device=DEVICE,
        verbose=0,
    )

    reports = layerwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )

# ------------------------------------------------------------
# Step - 3. 根据逐层误差情况，将部分难以量化的层直接送到非量化平台
# 在这个例子中，我们解除前十个误差最大的层的量化，这只是一个示例
# 为了确保量化精度达到较高水准，通常只有个别层需要解除量化。
# 不要妄图单纯使用调度解决所有精度问题，调度体现了运行效率与网络精度之间的权衡
# 我们后续还可以通过调节校准算法与量化参数来提升精度
# ------------------------------------------------------------

# 从大到小排序单层误差
sensitivity = [(op_name, error) for op_name, error in reports.items()]
sensitivity = sorted(sensitivity, key=lambda x: x[1], reverse=True)

# 将前十个误差最大的层送上 FP32
for op_name, _ in sensitivity[:10]:
    QS.dispatching_table.append(operation=op_name, platform=TargetPlatform.FP32)

# ------------------------------------------------------------
# Step - 4. 选择一个合适的校准算法，最小化量化误差
# 这一过程需要你手动调整 QS 中的校准算法，需要反复执行和对比
# 在这里我们只推荐你更改激活值的校准算法，对于参数而言
# 在 INT8 的量化场景中，minmax 往往都是最好的校准策略
# 在这个场景下，不能使用 isotone 策略
# ------------------------------------------------------------
for calib_algo in {'minmax', 'percentile', 'kl', 'mse'}:
    QS.quantize_activation_setting.calib_algorithm = calib_algo

    with ENABLE_CUDA_KERNEL():
        quantized = quantize_torch_model(
            model=model,
            calib_dataloader=CALIBRATION,
            calib_steps=32,
            input_shape=INPUT_SHAPE,
            collate_fn=collate_fn,
            platform=PLATFORM,
            setting=QS,
            onnx_export_file='Output/onnx.model',
            device=DEVICE,
            verbose=0,
        )

        print(f'Error Report of Algorithm {calib_algo}: ')
        reports = graphwise_error_analyse(
            graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
        )

# 在确定了一种校准算法之后，你还可以修改 esp_ppq.core.common 中的相关属性来取得更优结果
# 下列参数将影响校准效果：
# Observer 中 hist 箱子的个数
# OBSERVER_KL_HIST_BINS = 4096
# Observer 中 percentile 的参数
# OBSERVER_PERCENTILE = 0.9999
# Observer 中 mse 校准方法 hist 箱子的个数
# OBSERVER_MSE_HIST_BINS = 2048
# Observer 中 mse 计算的间隔，间隔越小，所需时间越长
# OBSERVER_MSE_COMPUTE_INTERVAL = 8

# 在完成测试后，选择一种最为合适的校准算法，我们此处以 percentile 为例
QS.quantize_activation_setting.calib_algorithm = 'percentile'

# ------------------------------------------------------------
# Step - 5. 再次检查我们的量化误差，如果与预期仍有差距
# 则我们可以进一步使用优化过程来调节网络参数
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    # 首先我们调节 equalization 算法的参数
    # 调节时关闭 lsq_optimization 以缩短流程执行时间
    QS.equalization = True  # 试试 True 或 False
    QS.equalization_setting.opt_level = 1  # 试试 1 或 2
    QS.equalization_setting.iterations = 10  # 试试 3, 10, 100
    QS.equalization_setting.value_threshold = 0.5  # 试试 0, 0.5, 2

    # 之后我们调节 LSQ 算法的参数
    QS.lsq_optimization = True
    QS.lsq_optimization_setting.block_size = 4  # 试试 1, 4, 6
    QS.lsq_optimization_setting.collecting_device = 'cuda'  # 如果显存不够你就写 cpu
    QS.lsq_optimization_setting.is_scale_trainable = True  # 试试 True 或者 False
    QS.lsq_optimization_setting.lr = 1e-5  # 试试 1e-5, 3e-5, 1e-4
    QS.lsq_optimization_setting.steps = 500  # 试试 300, 500, 2000

    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        setting=QS,
        onnx_export_file='Output/onnx.model',
        device=DEVICE,
        verbose=0,
    )

    reports = graphwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )


# ------------------------------------------------------------
# Step - 6. 最后，我们导出模型，并在 onnxruntime 上完成校验
# ------------------------------------------------------------
ONNX_OUTPUT_PATH = 'Output/model.onnx'
executor, reference_outputs = TorchExecutor(quantized), []
for sample in CALIBRATION:
    reference_outputs.append(executor.forward(collate_fn(sample)))

export_ppq_graph(graph=quantized, platform=TargetPlatform.ONNXRUNTIME, graph_save_to=ONNX_OUTPUT_PATH)

try:
    import onnxruntime
except ImportError as e:
    raise Exception('Onnxruntime is not installed.')

sess = onnxruntime.InferenceSession(ONNX_OUTPUT_PATH, providers=['CUDAExecutionProvider'])
onnxruntime_outputs = []
for sample in CALIBRATION:
    onnxruntime_outputs.append(
        sess.run(
            output_names=[name for name in quantized.outputs], input_feed={'input.1': convert_any_to_numpy(sample)}
        )
    )

name_of_output = [name for name in quantized.outputs]
for oidx, output in enumerate(name_of_output):
    y_pred, y_real = [], []
    for reference_output, onnxruntime_output in zip(reference_outputs, onnxruntime_outputs):
        y_pred.append(convert_any_to_torch_tensor(reference_output[oidx], device='cpu').unsqueeze(0))
        y_real.append(convert_any_to_torch_tensor(onnxruntime_output[oidx], device='cpu').unsqueeze(0))
    y_pred = torch.cat(y_pred, dim=0)
    y_real = torch.cat(y_real, dim=0)
    print(f'Simulating Error For {output}: {torch_snr_error(y_pred=y_pred, y_real=y_real).item():.4f}')



import sys
import time
from enum import Enum
from typing import Union

R_BEGIN = '\033[38;5;1m'
G_BEGIN = '\033[38;5;2m'
Y_BEGIN = '\033[38;5;3m'

COLOR_END = ' \033[m'

initialized_loggers = {}


def get_current_time():
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())


class LEVEL(Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3

    def __le__(self, other):
        return self.value <= other.value

    @staticmethod
    def convert(level: str):
        level = level.lower()
        if level == 'debug':
            return LEVEL.DEBUG
        elif level == 'info':
            return LEVEL.INFO
        elif level == 'warning':
            return LEVEL.WARNING
        elif level == 'error':
            return LEVEL.ERROR
        else:
            raise ValueError(
                f'the lowercase of param level should be one of debug/info/warning/errorhowever {level} is given'
            )


class Handler:
    def __init__(self, file_name: str = None, level: LEVEL = LEVEL.INFO) -> None:
        self._file_name = file_name
        self._level = level
        self._fd = sys.stdout if file_name is None else open(file_name, 'w+', encoding='utf-8')

    def process(self, msg: str, level: LEVEL) -> None:
        if self._level <= level:
            self._fd.write('\r' + msg + '\n')
            self._fd.flush()

    def set_level(self, level: LEVEL) -> None:
        self._level = level


class NaiveLogger(object):
    """A very naive implementation of colored logger, but would be suffice in
    single-process situation here where no race condition would happen."""

    __create_key = object()

    def __init__(self, create_key: object, name: str, level: LEVEL, file_name: str) -> None:
        assert create_key == NaiveLogger.__create_key, 'logger instance must be created using NaiveLogger.get_logger'
        self._name = name
        self._level = level
        self._handlers = {'stdout': Handler(level=level)}
        if file_name is not None:
            self.register_handler(file_name)

    def set_level(self, level: Union[str, LEVEL]):
        if isinstance(level, str):
            level = LEVEL.convert(level)
        assert isinstance(level, LEVEL), 'level should be given by either str or LEVEL instances'
        self._level = level
        for handler in self._handlers.values():
            handler.set_level(level)

    @classmethod
    def get_logger(cls, name: str, level: Union[str, LEVEL] = LEVEL.INFO, file_name: str = None):
        if name in initialized_loggers:
            return initialized_loggers[name]

        if isinstance(level, str):
            level = LEVEL.convert(level)

        assert isinstance(level, LEVEL), 'level should be given by either str or LEVEL instances'
        logger = NaiveLogger(cls.__create_key, name, level, file_name)
        initialized_loggers[name] = logger
        return logger

    def wrap_header(self, type: str) -> str:
        cur_time = get_current_time()
        return '[{}][{}][{}]: '.format(type, self._name, cur_time)

    def info(self, msg: str):
        header = self.wrap_header('INFO')
        print_msg = G_BEGIN + header + COLOR_END + msg

        for handler in self._handlers.values():
            if handler._file_name is not None:
                handler.process(msg, LEVEL.INFO)
            else:
                handler.process(print_msg, LEVEL.INFO)

    def warning(self, msg: str):
        header = self.wrap_header('WARNING')
        print_msg = Y_BEGIN + header + COLOR_END + msg

        for handler in self._handlers.values():
            if handler._file_name is not None:
                handler.process(msg, LEVEL.WARNING)
            else:
                handler.process(print_msg, LEVEL.WARNING)

    def error(self, msg: str):
        header = self.wrap_header('ERROR')
        print_msg = R_BEGIN + header + COLOR_END + msg

        for handler in self._handlers.values():
            if handler._file_name is not None:
                handler.process(msg, LEVEL.ERROR)
            else:
                handler.process(print_msg, LEVEL.ERROR)

    def debug(self, msg: str):
        header = self.wrap_header('DEBUG')
        print_msg = G_BEGIN + header + COLOR_END + msg

        for handler in self._handlers.values():
            if handler._file_name is not None:
                handler.process(msg, LEVEL.DEBUG)
            else:
                handler.process(print_msg, LEVEL.DEBUG)

    def register_handler(self, file_name: str, level: Union[str, LEVEL] = LEVEL.INFO):
        if file_name not in self._handlers:
            if isinstance(level, str):
                level = LEVEL.convert(level)
            assert isinstance(level, LEVEL), 'level should be given by either str or LEVEL instances'
            self._handlers[file_name] = Handler(file_name, level)

    def remove_handler(self, file_name: str):
        if file_name in self._handlers:
            handler = self._handlers[file_name]
            if handler._file_name is not None:
                handler._fd.close()
            self._handlers.pop(file_name)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

class TypeInfoValue(object):
    NONE = 0
    tensor_type = 1
    sequence_type = 2
    map_type = 3
    optional_type = 4

def TypeInfoValueCreator(unionType, table):
    from flatbuffers.table import Table
    if not isinstance(table, Table):
        return None
    if unionType == TypeInfoValue().tensor_type:
        import FlatBuffers.Dl.TensorTypeAndShape
        return FlatBuffers.Dl.TensorTypeAndShape.TensorTypeAndShapeT.InitFromBuf(table.Bytes, table.Pos)
    if unionType == TypeInfoValue().sequence_type:
        import FlatBuffers.Dl.SequenceType
        return FlatBuffers.Dl.SequenceType.SequenceTypeT.InitFromBuf(table.Bytes, table.Pos)
    if unionType == TypeInfoValue().map_type:
        import FlatBuffers.Dl.MapType
        return FlatBuffers.Dl.MapType.MapTypeT.InitFromBuf(table.Bytes, table.Pos)
    if unionType == TypeInfoValue().optional_type:
        import FlatBuffers.Dl.OptionalType
        return FlatBuffers.Dl.OptionalType.OptionalTypeT.InitFromBuf(table.Bytes, table.Pos)
    return None



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
import os
import sys
import time
from collections import defaultdict
from typing import Callable, Iterable

import pycuda.driver as cuda
import tensorrt as trt
import torch
from tqdm import tqdm

from esp_ppq import TorchExecutor, convert_any_to_numpy, convert_any_to_torch_tensor, torch_snr_error
from esp_ppq.api import ENABLE_CUDA_KERNEL
from esp_ppq.IR import BaseGraph
from esp_ppq.quantization.analyse.util import MeasurePrinter

logging.basicConfig(level=logging.INFO)
logging.getLogger("EngineBuilder").setLevel(logging.INFO)
log = logging.getLogger("EngineBuilder")


# Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


# Allocates all buffers required for an engine, i.e. host/device inputs/outputs.
def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


# This function is generalized for multiple inputs/outputs.
# inputs and outputs are expected to be lists of HostDeviceMem objects.
def do_inference(context, bindings, inputs, outputs, stream):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


# This function is generalized for multiple inputs/outputs for full dimension networks.
# inputs and outputs are expected to be lists of HostDeviceMem objects.
def do_inference_v2(context, bindings, inputs, outputs, stream):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


class EngineBuilder:
    """
    Parses an ONNX graph and builds a TensorRT engine from it.
    """

    def __init__(self, verbose=False):
        """
        :param verbose: If enabled, a higher verbosity level will be set on the TensorRT logger.
        """
        self.trt_logger = trt.Logger(trt.Logger.INFO)
        if verbose:
            self.trt_logger.min_severity = trt.Logger.Severity.VERBOSE

        trt.init_libnvinfer_plugins(self.trt_logger, namespace="")

        self.builder = trt.Builder(self.trt_logger)
        self.config = self.builder.create_builder_config()
        self.config.max_workspace_size = 8 * (2**30)  # 8 GB

        self.batch_size = None
        self.network = None
        self.parser = None

    def create_network(self, onnx_path):
        """
        Parse the ONNX graph and create the corresponding TensorRT network definition.
        :param onnx_path: The path to the ONNX graph to load.
        """
        network_flags = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)

        self.network = self.builder.create_network(network_flags)
        self.parser = trt.OnnxParser(self.network, self.trt_logger)

        onnx_path = os.path.realpath(onnx_path)
        with open(onnx_path, "rb") as f:
            if not self.parser.parse(f.read()):
                log.error("Failed to load ONNX file: {}".format(onnx_path))
                for error in range(self.parser.num_errors):
                    log.error(self.parser.get_error(error))
                sys.exit(1)

        inputs = [self.network.get_input(i) for i in range(self.network.num_inputs)]
        outputs = [self.network.get_output(i) for i in range(self.network.num_outputs)]

        log.info("Network Description")
        for input in inputs:
            self.batch_size = input.shape[0]
            log.info("Input '{}' with shape {} and dtype {}".format(input.name, input.shape, input.dtype))
        for output in outputs:
            log.info("Output '{}' with shape {} and dtype {}".format(output.name, output.shape, output.dtype))
        assert self.batch_size > 0
        self.builder.max_batch_size = self.batch_size

    def create_engine(
        self,
        engine_path,
        precision,
        calib_input=None,
        calib_cache=None,
        calib_num_images=25000,
        calib_batch_size=8,
        calib_preprocessor=None,
    ):
        """
        Build the TensorRT engine and serialize it to disk.
        :param engine_path: The path where to serialize the engine to.
        :param precision: The datatype to use for the engine, either 'fp32', 'fp16' or 'int8'.
        :param calib_input: The path to a directory holding the calibration images.
        :param calib_cache: The path where to write the calibration cache to, or if it already exists, load it from.
        :param calib_num_images: The maximum number of images to use for calibration.
        :param calib_batch_size: The batch size to use for the calibration process.
        :param calib_preprocessor: The ImageBatcher preprocessor algorithm to use.
        """
        engine_path = os.path.realpath(engine_path)
        engine_dir = os.path.dirname(engine_path)
        os.makedirs(engine_dir, exist_ok=True)
        log.info("Building {} Engine in {}".format(precision, engine_path))

        inputs = [self.network.get_input(i) for i in range(self.network.num_inputs)]

        if precision == "fp16":
            if not self.builder.platform_has_fast_fp16:
                log.warning("FP16 is not supported natively on this platform/device")
            else:
                self.config.set_flag(trt.BuilderFlag.FP16)

        with self.builder.build_engine(self.network, self.config) as engine, open(engine_path, "wb") as f:
            log.info("Serializing engine to file: {:}".format(engine_path))
            f.write(engine.serialize())


def setDynamicRange(network, json_file: str):
    """Sets ranges for network layers."""
    with open(json_file) as file:
        quant_param_json = json.load(file)
    act_quant = quant_param_json["act_quant_info"]

    for i in range(network.num_inputs):
        input_tensor = network.get_input(i)
        if act_quant.__contains__(input_tensor.name):
            value = act_quant[input_tensor.name]
            tensor_max = abs(value)
            tensor_min = -abs(value)
            input_tensor.dynamic_range = (tensor_min, tensor_max)

    for i in range(network.num_layers):
        layer = network.get_layer(i)

        for output_index in range(layer.num_outputs):
            tensor = layer.get_output(output_index)

            if act_quant.__contains__(tensor.name):
                value = act_quant[tensor.name]
                tensor_max = abs(value)
                tensor_min = -abs(value)
                tensor.dynamic_range = (tensor_min, tensor_max)


def build_engine(
    onnx_file: str,
    engine_file: str,
    fp16: bool = True,
    int8: bool = False,
    int8_scale_file: str = None,
    explicit_batch: bool = True,
    workspace: int = 4294967296,  # 4GB
):
    TRT_LOGGER = trt.Logger()
    """
    Build a TensorRT Engine with given onnx model.

    Flag int8, fp16 specifies the precision of layer:
        For building FP32 engine: set int8 = False, fp16 = False, int8_scale_file = None
        For building FP16 engine: set int8 = False, fp16 = True, int8_scale_file = None
        For building INT8 engine: set int8 = True, fp16 = True, int8_scale_file = 'json file name'

    """

    if int8 is True:
        if int8_scale_file is None:
            raise ValueError(
                'Build Quantized TensorRT Engine Requires a JSON file which specifies variable scales, '
                'however int8_scale_file is None now.'
            )

    builder = trt.Builder(TRT_LOGGER)
    if explicit_batch:
        network = builder.create_network(1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH))
    else:
        network = builder.create_network()

    config = builder.create_builder_config()

    parser = trt.OnnxParser(network, TRT_LOGGER)
    config.max_workspace_size = workspace

    if not os.path.exists(onnx_file):
        raise FileNotFoundError(f'ONNX file {onnx_file} not found')

    with open(onnx_file, 'rb') as model:
        if not parser.parse(model.read()):
            print('ERROR: Failed to parse the ONNX file.')
            for error in range(parser.num_errors):
                print(parser.get_error(error))
            return None

    if fp16:
        config.set_flag(trt.BuilderFlag.FP16)
    if int8_scale_file is not None and int8:
        config.set_flag(trt.BuilderFlag.INT8)
        setDynamicRange(network, int8_scale_file)

    engine = builder.build_engine(network, config)

    with open(engine_file, "wb") as f:
        f.write(engine.serialize())


class MyProfiler(trt.IProfiler):
    def __init__(self, steps: int):
        trt.IProfiler.__init__(self)
        self.total_runtime = 0.0
        self.recorder = defaultdict(lambda: 0.0)
        self.steps = steps

    def report_layer_time(self, layer_name: str, ms: float):
        self.total_runtime += ms * 1000 / self.steps
        self.recorder[layer_name] += ms * 1000 / self.steps

    def report(self):
        MeasurePrinter(data=self.recorder, measure='RUNNING TIME(us)', order=None).print()
        print(f'\nTotal Inference Time: {self.total_runtime:.4f}(us)')
        print('You should notice that inference time != Lantancy, cause layer can be executed concurrently.')


def Benchmark(engine_file: str, steps: int = 10000) -> float:
    """Benckmark TensorRT Engine"""
    logger = trt.Logger(trt.Logger.INFO)
    import pycuda.autoinit

    with open(engine_file, 'rb') as f, trt.Runtime(logger) as runtime:
        engine = runtime.deserialize_cuda_engine(f.read())

    with engine.create_execution_context() as context:
        inputs, outputs, bindings, stream = allocate_buffers(context.engine)
        for input in inputs:
            input.host = convert_any_to_numpy(torch.rand(input.host.shape).float())

        tick = time.time()
        for _ in tqdm(range(steps), desc='TensorRT is running...'):
            do_inference_v2(context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)
        tok = time.time()
        print(f'Time span: {tok - tick: .4f} sec')
    return tick - tok


def Profiling(engine_file: str, steps: int = 1000):
    """Profiling TensorRT Engine"""
    logger = trt.Logger(trt.Logger.ERROR)
    import pycuda.autoinit

    with open(engine_file, 'rb') as f, trt.Runtime(logger) as runtime:
        engine = runtime.deserialize_cuda_engine(f.read())

    with engine.create_execution_context() as context:
        context.profiler = MyProfiler(steps)
        inputs, outputs, bindings, stream = allocate_buffers(context.engine)

        for input in inputs:
            input.host = convert_any_to_numpy(torch.rand(input.host.shape).float())

        for _ in tqdm(range(steps), desc='TensorRT is running...'):
            do_inference_v2(context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)
        context.profiler.report()


def TestAlignment(engine_file: str, graph: BaseGraph, samples: Iterable, collate_fn: Callable = None) -> dict:
    """Test Alignment with TensorRT Engine and PPQ Graph."""
    logger = trt.Logger(trt.Logger.ERROR)

    feed_dicts = []
    for sample in samples:
        if collate_fn is not None:
            sample = collate_fn(sample)
        feed_dict = {}
        if isinstance(sample, torch.Tensor):
            assert len(graph.inputs) == 1, 'Graph Needs More than 1 input tensor, however only 1 was given.'
            for name in graph.inputs:
                feed_dict[name] = sample
        elif isinstance(sample, list):
            for name, value in zip(graph.inputs, sample):
                feed_dict[name] = value
        elif isinstance(sample, dict):
            feed_dict = sample
        else:
            raise TypeError('Given Sample is Invalid.')
        feed_dicts.append(feed_dict)

    TensorRT_Results, PPQ_Results = [], []
    with ENABLE_CUDA_KERNEL():
        executor = TorchExecutor(graph)
        for feed_dict in tqdm(feed_dicts, desc='PPQ Infer...'):
            PPQ_Results.append([value.cpu() for value in executor.forward(feed_dict)])

    import pycuda.autoinit

    with open(engine_file, 'rb') as f, trt.Runtime(logger) as runtime:
        engine = runtime.deserialize_cuda_engine(f.read())

    with engine.create_execution_context() as context:
        inputs, outputs, bindings, stream = allocate_buffers(context.engine)

        for feed_dict in tqdm(feed_dicts, desc='TensorRT Infer...'):
            for name, input in zip(graph.inputs, inputs):
                input.host = convert_any_to_numpy(feed_dict[name])

            results = do_inference_v2(context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)

            TensorRT_Results.append([convert_any_to_torch_tensor(value) for value in results])

    collector = {}
    for ref, pred in zip(TensorRT_Results, PPQ_Results):
        for name in graph.outputs:
            collector[name] = 0.0
        for name, ref_, pred_ in zip(graph.outputs, ref, pred):
            collector[name] += torch_snr_error(pred_.reshape([1, -1]), ref_.reshape([1, -1])).item()

    for name, value in collector.items():
        collector[name] = value / len(samples)
    return collector
